import unittest
import os.path, random
from shutil import rmtree

from whoosh import fields, query
from whoosh.filedb.filestore import FileStorage, RamStorage
from whoosh.filedb.filewriting import NO_MERGE
from whoosh.util import length_to_byte, byte_to_length, permutations
from whoosh.writing import BatchWriter, IndexingError


class TestIndexing(unittest.TestCase):
    def make_index(self, dirname, schema, ixname):
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        st = FileStorage(dirname)
        ix = st.create_index(schema, indexname=ixname)
        return ix
    
    def destroy_index(self, dirname):
        if os.path.exists(dirname):
            try:
                rmtree(dirname)
            except OSError, e:
                pass
    
    def test_creation(self):
        s = fields.Schema(content=fields.TEXT(phrase=True),
                          title=fields.TEXT(stored=True),
                          path=fields.ID(stored=True),
                          tags=fields.KEYWORD(stored=True),
                          quick=fields.NGRAM,
                          note=fields.STORED)
        st = RamStorage()
        
        ix = st.create_index(s)
        w = ix.writer()
        w.add_document(title=u"First", content=u"This is the first document", path=u"/a",
                       tags=u"first second third", quick=u"First document", note=u"This is the first document")
        w.add_document(content=u"Let's try this again", title=u"Second", path=u"/b",
                       tags=u"Uno Dos Tres", quick=u"Second document", note=u"This is the second document")
        w.commit()
    
    def test_empty_commit(self):
        s = fields.Schema(id=fields.ID(stored=True))
        ix = self.make_index("testindex", s, "emptycommit")
        w = ix.writer()
        w.add_document(id=u"1")
        w.add_document(id=u"2")
        w.add_document(id=u"3")
        w.commit()
        
        w = ix.writer()
        w.commit()
    
    def test_multipool(self):
        try:
            import multiprocessing
        except ImportError:
            return
        
        domain = (u"alfa", u"bravo", u"charlie", u"delta", u"echo", u"foxtrot", u"golf",
                  u"hotel", u"india", u"juliet", u"kilo", u"lima", u"mike", u"november")
        
        s = fields.Schema(content=fields.TEXT, id=fields.ID)
        ix = self.make_index("testindex", s, "multipool")
        
        w = ix.writer(procs=4)
        for _ in xrange(1000):
            w.add_document(content=u" ".join(random.sample(domain, 5)),
                           id=random.choice(domain))
        w.commit()
        
        ix.close()
        self.destroy_index("testindex")
    
    def test_integrity(self):
        s = fields.Schema(name=fields.TEXT, value=fields.TEXT)
        st = RamStorage()
        ix = st.create_index(s)
        
        w = ix.writer()
        w.add_document(name=u"Yellow brown", value=u"Blue red green purple?")
        w.add_document(name=u"Alpha beta", value=u"Gamma delta epsilon omega.")
        w.commit()
        
        w = ix.writer()
        w.add_document(name=u"One two", value=u"Three four five.")
        w.commit()
        
        tr = ix.reader()
        self.assertEqual(ix.doc_count_all(), 3)
        self.assertEqual(list(tr.lexicon("name")), ["alpha", "beta", "brown", "one", "two", "yellow"])
    
    def test_lengths(self):
        s = fields.Schema(f1=fields.KEYWORD(stored=True, scorable=True),
                          f2=fields.KEYWORD(stored=True, scorable=True))
        ix = self.make_index("testindex", s, "test_lengths")
        
        try:
            w = ix.writer()
            try:
                tokens = u"ABCDEFG"
                from itertools import cycle, islice
                lengths = [10, 20, 2, 102, 45, 3, 420, 2]
                for length in lengths:
                    w.add_document(f2=u" ".join(islice(cycle(tokens), length)))
                w.commit()
            except Exception:
                w.cancel()
                raise
            
            dr = ix.reader()
            try:
                ls1 = [dr.doc_field_length(i, "f1") for i in xrange(0, len(lengths))]
                self.assertEqual(ls1, [0] * len(lengths))
                ls2 = [dr.doc_field_length(i, "f2") for i in xrange(0, len(lengths))]
                self.assertEqual(ls2, [byte_to_length(length_to_byte(l))
                                       for l in lengths])
            finally:
                dr.close()
                
        finally:
            ix.close()
            self.destroy_index("testindex")
    
    def test_lengths_ram(self):
        s = fields.Schema(f1=fields.KEYWORD(stored=True, scorable=True),
                          f2=fields.KEYWORD(stored=True, scorable=True))
        st = RamStorage()
        ix = st.create_index(s)
        w = ix.writer()
        w.add_document(f1=u"A B C D E", f2=u"X Y Z")
        w.add_document(f1=u"B B B B C D D Q", f2=u"Q R S T")
        w.add_document(f1=u"D E F", f2=u"U V A B C D E")
        w.commit()
        
        dr = ix.reader()
        self.assertEqual(dr.stored_fields(0)["f1"], "A B C D E")
        self.assertEqual(dr.doc_field_length(0, "f1"), 5)
        self.assertEqual(dr.doc_field_length(1, "f1"), 8)
        self.assertEqual(dr.doc_field_length(2, "f1"), 3)
        self.assertEqual(dr.doc_field_length(0, "f2"), 3)
        self.assertEqual(dr.doc_field_length(1, "f2"), 4)
        self.assertEqual(dr.doc_field_length(2, "f2"), 7)
        
        self.assertEqual(dr.field_length("f1"), 16)
        self.assertEqual(dr.field_length("f2"), 14)
        self.assertEqual(dr.max_field_length("f1"), 8)
        self.assertEqual(dr.max_field_length("f2"), 7)
        
    def test_merged_lengths(self):
        s = fields.Schema(f1=fields.KEYWORD(stored=True, scorable=True),
                          f2=fields.KEYWORD(stored=True, scorable=True))
        st = RamStorage()
        ix = st.create_index(s)
        w = ix.writer()
        w.add_document(f1=u"A B C", f2=u"X")
        w.add_document(f1=u"B C D E", f2=u"Y Z")
        w.commit()
        
        w = ix.writer()
        w.add_document(f1=u"A", f2=u"B C D E X Y")
        w.add_document(f1=u"B C", f2=u"X")
        w.commit(NO_MERGE)
        
        w = ix.writer()
        w.add_document(f1=u"A B X Y Z", f2=u"B C")
        w.add_document(f1=u"Y X", f2=u"A B")
        w.commit(NO_MERGE)
        
        dr = ix.reader()
        self.assertEqual(dr.stored_fields(0)["f1"], u"A B C")
        self.assertEqual(dr.doc_field_length(0, "f1"), 3)
        self.assertEqual(dr.doc_field_length(2, "f2"), 6)
        self.assertEqual(dr.doc_field_length(4, "f1"), 5)
        dr.close()
        
    def test_frequency_keyword(self):
        s = fields.Schema(content=fields.KEYWORD)
        st = RamStorage()
        ix = st.create_index(s)
        
        w = ix.writer()
        w.add_document(content=u"A B C D E")
        w.add_document(content=u"B B B B C D D")
        w.add_document(content=u"D E F")
        w.commit()
        
        tr = ix.reader()
        self.assertEqual(tr.doc_frequency("content", u"B"), 2)
        self.assertEqual(tr.frequency("content", u"B"), 5)
        self.assertEqual(tr.doc_frequency("content", u"E"), 2)
        self.assertEqual(tr.frequency("content", u"E"), 2)
        self.assertEqual(tr.doc_frequency("content", u"A"), 1)
        self.assertEqual(tr.frequency("content", u"A"), 1)
        self.assertEqual(tr.doc_frequency("content", u"D"), 3)
        self.assertEqual(tr.frequency("content", u"D"), 4)
        self.assertEqual(tr.doc_frequency("content", u"F"), 1)
        self.assertEqual(tr.frequency("content", u"F"), 1)
        self.assertEqual(tr.doc_frequency("content", u"Z"), 0)
        self.assertEqual(tr.frequency("content", u"Z"), 0)
        self.assertEqual(list(tr), [("content", u"A", 1, 1),
                                    ("content", u"B", 2, 5),
                                    ("content", u"C", 2, 2),
                                    ("content", u"D", 3, 4),
                                    ("content", u"E", 2, 2),
                                    ("content", u"F", 1, 1)])
        tr.close()
        
    def test_frequency_text(self):
        s = fields.Schema(content=fields.KEYWORD)
        st = RamStorage()
        ix = st.create_index(s)
        
        w = ix.writer()
        w.add_document(content=u"alfa bravo charlie delta echo")
        w.add_document(content=u"bravo bravo bravo bravo charlie delta delta")
        w.add_document(content=u"delta echo foxtrot")
        w.commit()
        
        tr = ix.reader()
        self.assertEqual(tr.doc_frequency("content", u"bravo"), 2)
        self.assertEqual(tr.frequency("content", u"bravo"), 5)
        self.assertEqual(tr.doc_frequency("content", u"echo"), 2)
        self.assertEqual(tr.frequency("content", u"echo"), 2)
        self.assertEqual(tr.doc_frequency("content", u"alfa"), 1)
        self.assertEqual(tr.frequency("content", u"alfa"), 1)
        self.assertEqual(tr.doc_frequency("content", u"delta"), 3)
        self.assertEqual(tr.frequency("content", u"delta"), 4)
        self.assertEqual(tr.doc_frequency("content", u"foxtrot"), 1)
        self.assertEqual(tr.frequency("content", u"foxtrot"), 1)
        self.assertEqual(tr.doc_frequency("content", u"zulu"), 0)
        self.assertEqual(tr.frequency("content", u"zulu"), 0)
        self.assertEqual(list(tr), [("content", u"alfa", 1, 1),
                                    ("content", u"bravo", 2, 5),
                                    ("content", u"charlie", 2, 2),
                                    ("content", u"delta", 3, 4),
                                    ("content", u"echo", 2, 2),
                                    ("content", u"foxtrot", 1, 1)])
        tr.close()
    
    def test_deletion(self):
        s = fields.Schema(key=fields.ID, name=fields.TEXT, value=fields.TEXT)
        ix = RamStorage().create_index(s)
        
        w = ix.writer()
        w.add_document(key=u"A", name=u"Yellow brown", value=u"Blue red green purple?")
        w.add_document(key=u"B", name=u"Alpha beta", value=u"Gamma delta epsilon omega.")
        w.add_document(key=u"C", name=u"One two", value=u"Three four five.")
        w.commit()
        
        w = ix.writer()
        count = w.delete_by_term("key", u"B")
        self.assertEqual(count, 1)
        w.commit(merge=False)
        
        self.assertEqual(ix.doc_count_all(), 3)
        self.assertEqual(ix.doc_count(), 2)
        
        w = ix.writer()
        w.add_document(key=u"A", name=u"Yellow brown", value=u"Blue red green purple?")
        w.add_document(key=u"B", name=u"Alpha beta", value=u"Gamma delta epsilon omega.")
        w.add_document(key=u"C", name=u"One two", value=u"Three four five.")
        w.commit()
        
        # This will match both documents with key == B, one of which is already
        # deleted. This should not raise an error.
        w = ix.writer()
        count = w.delete_by_term("key", u"B")
        self.assertEqual(count, 1)
        w.commit()
        
        ix.optimize()
        self.assertEqual(ix.doc_count_all(), 4)
        self.assertEqual(ix.doc_count(), 4)
        tr = ix.reader()
        self.assertEqual(list(tr.lexicon("name")), ["brown", "one", "two", "yellow"])
        tr.close()

    def test_writer_reuse(self):
        s = fields.Schema(key=fields.ID)
        ix = RamStorage().create_index(s)
        
        w = ix.writer()
        w.add_document(key=u"A")
        w.add_document(key=u"B")
        w.add_document(key=u"C")
        w.commit()
        
        # You can't re-use a commited/canceled writer
        self.assertRaises(IndexingError, w.add_document, key=u"D")
        self.assertRaises(IndexingError, w.update_document, key=u"B")
        self.assertRaises(IndexingError, w.delete_document, 0)
        self.assertRaises(IndexingError, w.add_reader, None)
        self.assertRaises(IndexingError, w.add_field, "name", fields.ID)
        self.assertRaises(IndexingError, w.remove_field, "key")
        self.assertRaises(IndexingError, w.searcher)

    def test_update(self):
        # Test update with multiple unique keys
        SAMPLE_DOCS = [{"id": u"test1", "path": u"/test/1", "text": u"Hello"},
                       {"id": u"test2", "path": u"/test/2", "text": u"There"},
                       {"id": u"test3", "path": u"/test/3", "text": u"Reader"},
                       ]
        
        schema = fields.Schema(id=fields.ID(unique=True, stored=True),
                               path=fields.ID(unique=True, stored=True),
                               text=fields.TEXT)
        ix = self.make_index("testindex", schema, "test_update")
        writer = ix.writer()
        for doc in SAMPLE_DOCS:
            writer.add_document(**doc)
        writer.commit()
        
        writer = ix.writer()
        writer.update_document(id=u"test2", path=u"test/1", text=u"Replacement")
        writer.commit()
        
        ix.close()
        self.destroy_index("testindex")

    def test_reindex(self):
        SAMPLE_DOCS = [
            {'id': u'test1', 'text': u'This is a document. Awesome, is it not?'},
            {'id': u'test2', 'text': u'Another document. Astounding!'},
            {'id': u'test3', 'text': u'A fascinating article on the behavior of domestic steak knives.'},
        ]

        schema = fields.Schema(text=fields.TEXT(stored=True),
                               id=fields.ID(unique=True, stored=True))
        ix = self.make_index("testindex", schema, "test_reindex")
        
        def reindex():
            writer = ix.writer()
            try:
                for doc in SAMPLE_DOCS:
                    writer.update_document(**doc)
                writer.commit()
            except:
                writer.cancel()
                raise

        reindex()
        self.assertEqual(ix.doc_count_all(), 3)
        reindex()
        self.assertEqual(ix.doc_count_all(), 3)
        
        ix.close()
        self.destroy_index("testindex")
            
    def test_noscorables1(self):
        values = [u"alfa", u"bravo", u"charlie", u"delta", u"echo", u"foxtrot",
                  u"golf", u"hotel", u"india", u"juliet", u"kilo", u"lima"]
        from random import choice, sample, randint
        
        times = 1000
        
        schema = fields.Schema(id=fields.ID, tags=fields.KEYWORD)
        ix = self.make_index("testindex", schema, "noscorables1")
        
        w = ix.writer()
        for i in xrange(times):
            w.add_document(id=choice(values), tags=u" ".join(sample(values, randint(2, 7))))
        w.commit()
        
        s = ix.searcher()
        s.search(query.Term("id", "bravo"))
        s.close()
        ix.close()
        self.destroy_index("testindex")
            
    def test_noscorables2(self):
        schema = fields.Schema(field=fields.ID)
        ix = self.make_index("testindex", schema, "noscorables2")
        writer = ix.writer()
        writer.add_document(field=u'foo')
        writer.commit()
        ix.close()
        self.destroy_index("testindex")
    
    def test_multi(self):
        schema = fields.Schema(id=fields.ID(stored=True),
                               content=fields.KEYWORD(stored=True))
        ix =self.make_index("testindex", schema, "multi")
        
        writer = ix.writer()
        writer.add_document(id=u"1", content=u"alfa bravo charlie 0") #deleted 1
        writer.add_document(id=u"2", content=u"bravo charlie delta echo 1") #deleted 1
        writer.add_document(id=u"3", content=u"charlie delta echo foxtrot 2") #deleted 2
        writer.commit()
        
        writer = ix.writer()
        writer.delete_by_term("id", "1")
        writer.delete_by_term("id", "2")
        writer.add_document(id=u"1", content=u"apple bear cherry donut 3")
        writer.add_document(id=u"2", content=u"bear cherry donut eggs 4")
        writer.add_document(id=u"4", content=u"delta echo foxtrot golf 5") #deleted 2
        writer.add_document(id=u"5", content=u"echo foxtrot golf hotel 6")
        writer.commit(merge=False)
        
        writer = ix.writer()
        writer.delete_by_term("id", "3")
        writer.delete_by_term("id", "4")
        writer.add_document(id=u"3", content=u"cherry donut eggs falafel 7")
        writer.add_document(id=u"4", content=u"donut eggs falafel grape 8")
        writer.add_document(id=u"6", content=u" foxtrot golf hotel india 9")
        writer.commit(merge=False)

        self.assertEqual(ix.doc_count(), 6)
        
        s = ix.searcher()
        
        r = s.search(query.Prefix("content", u"d"), optimize=False)
        self.assertEqual(sorted([d["id"] for d in r]), ["1", "2", "3", "4"])
        
        r = s.search(query.Prefix("content", u"d"))
        self.assertEqual(sorted([d["id"] for d in r]), ["1", "2", "3", "4"])
        
        r = s.search(query.Prefix("content", u"d"), limit=None)
        self.assertEqual(sorted([d["id"] for d in r]), ["1", "2", "3", "4"])
        
    def test_deleteall(self):
        schema = fields.Schema(text=fields.TEXT)
        ix = RamStorage().create_index(schema)
        w = BatchWriter(ix, limit=10)
        domain = u"alfa bravo charlie delta echo".split()
        for ls in permutations(domain):
            w.add_document(text=u" ".join(ls))
        w.commit()
        
        # This is just a test, don't use this method to delete all docs IRL!
        doccount = ix.doc_count_all()
        w = ix.writer()
        for docnum in xrange(doccount):
            w.delete_document(docnum)
        w.commit()
        
        r = ix.searcher().search(query.Or([query.Term("text", u"alfa"), query.Term("text", u"bravo")]))
        self.assertEqual(len(r), 0)
        
        ix.optimize()
        self.assertEqual(ix.doc_count_all(), 0)
        self.assertEqual(list(ix.reader()), [])
        
        


if __name__ == '__main__':
    unittest.main()



import unittest

from whoosh import fields
from whoosh.filedb import pools, filestore


class FakeTermIndex(object):
    def __init__(self):
        self.d = {}
    
    def add(self, term, info):
        fieldname, text = term
        weight, offset, postcount = info
        self.d[term] = info
    
    def __getattr__(self, name):
        raise Exception("FTI name=%s" % name)
    

class FakePostWriter(object):
    def __init__(self):
        self.l = []
    
    def start(self, format):
        self.ps = []
        return len(self.l)
        
    def write(self, *args):
        self.ps.append(args)
    
    def finish(self):
        return len(self.ps)
    
    def __getattr__(self, name):
        raise Exception("FPW name=%s" % name)


class TestPool(unittest.TestCase):
    def test_addpostings(self):
        s = fields.Schema(text=fields.TEXT)
        st = filestore.RamStorage()
        
        p = pools.TempfilePool(s)
        try:
            p.add_posting("text", u"alfa", 0, 1.0, "\x00\x00\x00\x01")
            p.add_posting("text", u"bravo", 0, 2.0, "\x00\x00\x00\x02")
            p.add_posting("text", u"charlie", 0, 3.0, "\x00\x00\x00\x03")
            p.add_field_length(0, "text", 6)
            p.add_posting("text", u"bravo", 1, 4.0, "\x00\x00\x00\x04")
            p.add_posting("text", u"charlie", 1, 5.0, "\x00\x00\x00\x05")
            p.add_posting("text", u"delta", 1, 6.0, "\x00\x00\x00\x06")
            p.add_field_length(1, "text", 15)
            
            p.dump_run()
            
            doccount = 2
            lengthfile = st.create_file("test.len")
            termtable = FakeTermIndex()
            postwriter = FakePostWriter()
            
            p.finish(doccount, lengthfile, termtable, postwriter)
        finally:
            pass
            #p.cleanup()
        



if __name__ == '__main__':
    unittest.main()




import unittest

from whoosh.analysis import *
from whoosh.support.unicode import blocks, blockname, blocknum

class TestAnalysis(unittest.TestCase):
    def test_regextokenizer(self):
        value = u"AAAaaaBBBbbbCCCcccDDDddd"
        
        rex = RegexTokenizer("[A-Z]+")
        self.assertEqual([t.text for t in rex(value)],
                         [u"AAA", u"BBB", u"CCC", u"DDD"])
        
        rex = RegexTokenizer("[A-Z]+", gaps=True)
        self.assertEqual([t.text for t in rex(value)],
                         [u"aaa", u"bbb", u"ccc", u"ddd"])
    
    def test_composition1(self):
        ca = RegexTokenizer() | LowercaseFilter()
        self.assertEqual(ca.__class__.__name__, "CompositeAnalyzer")
        self.assertEqual(ca[0].__class__.__name__, "RegexTokenizer")
        self.assertEqual(ca[1].__class__.__name__, "LowercaseFilter")
        self.assertEqual([t.text for t in ca(u"ABC 123")], ["abc", "123"])
    
    def test_composition2(self):
        ca = RegexTokenizer() | LowercaseFilter()
        sa = ca | StopFilter()
        self.assertEqual(len(sa), 3)
        self.assertEqual(sa.__class__.__name__, "CompositeAnalyzer")
        self.assertEqual(sa[0].__class__.__name__, "RegexTokenizer")
        self.assertEqual(sa[1].__class__.__name__, "LowercaseFilter")
        self.assertEqual(sa[2].__class__.__name__, "StopFilter")
        self.assertEqual([t.text for t in sa(u"The ABC 123")], ["abc", "123"])
    
    def test_composition3(self):
        sa = RegexTokenizer() | StopFilter()
        self.assertEqual(sa.__class__.__name__, "CompositeAnalyzer")
    
    def test_filter_composition(self):
        filtersonly = LowercaseFilter() | StopFilter()
        generator = filtersonly(u"Hello there")
        self.assertRaises(AssertionError, list, generator)
        
        analyzer = RegexTokenizer() | filtersonly
        self.assertEqual([t.text for t in analyzer(u"The ABC 123")], ["abc", "123"])
        
    def test_composing_functions(self):
        def filter(tokens):
            for t in tokens:
                t.text = t.text.upper()
                yield t
                
        analyzer = RegexTokenizer() | filter
        self.assertEqual([t.text for t in analyzer(u"abc def")], ["ABC", "DEF"])
    
    def test_multifilter(self):
        f1 = LowercaseFilter()
        f2 = PassFilter()
        mf = MultiFilter(a=f1, b=f2)
        ana = RegexTokenizer(r"\S+") | mf
        text = u"ALFA BRAVO CHARLIE"
        self.assertEqual([t.text for t in ana(text, mode="a")], ["alfa", "bravo", "charlie"])
        self.assertEqual([t.text for t in ana(text, mode="b")], ["ALFA", "BRAVO", "CHARLIE"])
    
    def test_intraword(self):
        iwf = IntraWordFilter(mergewords=True, mergenums=True)
        ana = RegexTokenizer(r"\S+") | iwf
        
        def check(text, ls):
            self.assertEqual([(t.pos, t.text) for t in ana(text)], ls)
            
        check(u"PowerShot", [(0, "Power"), (1, "Shot"), (1, "PowerShot")])
        check(u"A's+B's&C's", [(0, "A"), (1, "B"), (2, "C"), (2, "ABC")])
        check(u"Super-Duper-XL500-42-AutoCoder!", [(0, "Super"), (1, "Duper"), (2, "XL"),
                                                   (2, "SuperDuperXL"), (3, "500"), (4, "42"),
                                                   (4, "50042"), (5, "Auto"), (6, "Coder"),
                                                   (6, "AutoCoder")])
    
    def test_biword(self):
        ana = RegexTokenizer(r"\w+") | BiWordFilter()
        result = [t.copy() for t in ana(u"the sign of four",
                                        chars=True, positions=True)]
        self.assertEqual(["the-sign", "sign-of", "of-four"],
                         [t.text for t in result])
        self.assertEqual([(0, 8), (4, 11), (9, 16)],
                         [(t.startchar, t.endchar) for t in result])
        self.assertEqual([0, 1, 2], [t.pos for t in result])
        
        result = [t.copy() for t in ana(u"single")]
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].text, "single")
        
    def test_unicode_blocks(self):
        self.assertEqual(blockname(u'a'), 'Basic Latin')
        self.assertEqual(blockname(unichr(0x0b80)), 'Tamil')
        self.assertEqual(blockname(unichr(2048)), None)
        self.assertEqual(blocknum(u'a'), 0)
        self.assertEqual(blocknum(unichr(0x0b80)), 22)
        self.assertEqual(blocknum(unichr(2048)), None)
        self.assertEqual(blocknum(u'a'), blocks.Basic_Latin)
        self.assertEqual(blocknum(unichr(0x0b80)), blocks.Tamil)
        
        

if __name__ == '__main__':
    unittest.main()



#===============================================================================
# Copyright 2007 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""Miscellaneous utility functions and classes.
"""

from array import array
from math import log
import codecs, re, sys, time

from collections import deque, defaultdict
from copy import copy
from functools import wraps
from struct import pack, unpack

from whoosh.system import IS_LITTLE


try:
    from itertools import permutations
except ImportError:
    # This function was only added to itertools in 2.6...
    def permutations(iterable, r=None):
        pool = tuple(iterable)
        n = len(pool)
        r = n if r is None else r
        if r > n:
            return
        indices = range(n)
        cycles = range(n, n-r, -1)
        yield tuple(pool[i] for i in indices[:r])
        while n:
            for i in reversed(range(r)):
                cycles[i] -= 1
                if cycles[i] == 0:
                    indices[i:] = indices[i+1:] + indices[i:i+1]
                    cycles[i] = n - i
                else:
                    j = cycles[i]
                    indices[i], indices[-j] = indices[-j], indices[i]
                    yield tuple(pool[i] for i in indices[:r])
                    break
            else:
                return


if sys.platform == 'win32':
    now = time.clock
else:
    now = time.time


# Note: these functions return a tuple of (text, length), so when you call
# them, you have to add [0] on the end, e.g. str = utf8encode(unicode)[0]

utf8encode = codecs.getencoder("utf_8")
utf8decode = codecs.getdecoder("utf_8")


# Functions

def array_to_string(a):
    if IS_LITTLE:
        a = copy(a)
        a.byteswap()
    return a.tostring()

def string_to_array(typecode, s):
    a = array(typecode)
    a.fromstring(s)
    if IS_LITTLE:
        a.byteswap()
    return a


def make_binary_tree(fn, args, **kwargs):
    """Takes a function/class that takes two positional arguments and a list of
    arguments and returns a binary tree of instances.
    
    >>> make_binary_tree(UnionMatcher, [matcher1, matcher2, matcher3])
    UnionMatcher(matcher1, UnionMatcher(matcher2, matcher3))
    
    Any keyword arguments given to this function are passed to the class
    initializer.
    """
    
    count = len(args)
    if not count:
        raise ValueError("Called make_binary_tree with empty list")
    elif count == 1:
        return args[0]
    
    half = count // 2
    return fn(make_binary_tree(fn, args[:half], **kwargs),
              make_binary_tree(fn, args[half:], **kwargs), **kwargs)


# Varint cache

# Build a cache of the varint byte sequences for the first N integers, so we
# don't have to constantly recalculate them on the fly. This makes a small but
# noticeable difference.

def _varint(i):
    s = ""
    while (i & ~0x7F) != 0:
        s += chr((i & 0x7F) | 0x80)
        i = i >> 7
    s += chr(i)
    return s

_varint_cache_size = 512
_varint_cache = []
for i in xrange(0, _varint_cache_size):
    _varint_cache.append(_varint(i))
_varint_cache = tuple(_varint_cache)

def varint(i):
    """Encodes the given integer into a string of the minimum number  of bytes.
    """
    if i < len(_varint_cache):
        return _varint_cache[i]
    return _varint(i)

def varint_to_int(vi):
    b = ord(vi[0])
    p = 1
    i = b & 0x7f
    shift = 7
    while b & 0x80 != 0:
        b = ord(vi[p])
        p += 1
        i |= (b & 0x7F) << shift
        shift += 7
    return i


def signed_varint(i):
    """Zig-zag encodes a signed integer into a varint.
    """
    
    if i >= 0:
        return varint(i << 1)
    return varint((i << 1) ^ (~0))

def decode_signed_varint(i):
    """Zig-zag decodes an integer value.
    """
    
    if not i & 1:
        return i >> 1
    return (i >> 1) ^ (~0)


def read_varint(readfn):
    """
    Reads a variable-length encoded integer.
    
    :param readfn: a callable that reads a given number of bytes,
        like file.read().
    """

    b = ord(readfn(1))
    i = b & 0x7F

    shift = 7
    while b & 0x80 != 0:
        b = ord(readfn(1))
        i |= (b & 0x7F) << shift
        shift += 7
    return i


_fib_cache = {}
def fib(n):
    """Returns the nth value in the Fibonacci sequence.
    """

    if n <= 2: return n
    if n in _fib_cache: return _fib_cache[n]
    result = fib(n - 1) + fib(n - 2)
    _fib_cache[n] = result
    return result


def float_to_byte(value, mantissabits=5, zeroexp=2):
    """Encodes a floating point number in a single byte.
    """

    # Assume int size == float size

    fzero = (63 - zeroexp) << mantissabits
    bits = unpack("i", pack("f", value))[0]
    smallfloat = bits >> (24 - mantissabits)
    if smallfloat < fzero:
        # Map negative numbers and 0 to 0
        # Map underflow to next smallest non-zero number
        if bits <= 0:
            return chr(0)
        else:
            return chr(1)
    elif smallfloat >= fzero + 0x100:
        # Map overflow to largest number
        return chr(255)
    else:
        return chr(smallfloat - fzero)

def byte_to_float(b, mantissabits=5, zeroexp=2):
    """Decodes a floating point number stored in a single byte.
    """
    b = ord(b)
    if b == 0:
        return 0.0

    bits = (b & 0xff) << (24 - mantissabits)
    bits += (63 - zeroexp) << 24
    return unpack("f", pack("i", bits))[0]


# Length-to-byte approximation functions

def length_to_byte(length):
    """Returns a logarithmic approximation of the given number, in the range
    0-255. The approximation has high precision at the low end (e.g.
    1 -> 0, 2 -> 1, 3 -> 2 ...) and low precision at the high end. Numbers
    equal to or greater than 108116 all approximate to 255.
    
    This is useful for storing field lengths, where the general case is small
    documents and very large documents are more rare.
    """
    
    # This encoding formula works up to 108116 -> 255, so if the length is
    # equal to or greater than that limit, just return 255.
    if length >= 108116: return 255
    
    # The parameters of this formula where chosen heuristically so that low
    # numbers would approximate closely, and the byte range 0-255 would cover
    # a decent range of document lengths (i.e. 1 to ~100000).
    return int(round(log((length/27.0)+1, 1.033)))

def _byte_to_length(n):
    return int(round((pow(1.033, n)-1)*27))

_length_byte_cache = array("i", (_byte_to_length(i) for i in xrange(256)))
byte_to_length = _length_byte_cache.__getitem__

# Prefix encoding functions

def first_diff(a, b):
    """Returns the position of the first differing character in the strings
    a and b. For example, first_diff('render', 'rending') == 4. This function
    limits the return value to 255 so the difference can be encoded in a single
    byte.
    """

    i = -1
    for i in xrange(0, len(a)):
        if a[i] != b[1]:
            return i
        if i == 255: return i


def prefix_encode(a, b):
    """Compresses string b as an integer (encoded in a byte) representing
    the prefix it shares with a, followed by the suffix encoded as UTF-8.
    """
    i = first_diff(a, b)
    return chr(i) + b[i:].encode("utf8")


def prefix_encode_all(ls):
    """Compresses the given list of (unicode) strings by storing each string
    (except the first one) as an integer (encoded in a byte) representing
    the prefix it shares with its predecessor, followed by the suffix encoded
    as UTF-8.
    """

    last = u''
    for w in ls:
        i = first_diff(last, w)
        yield chr(i) + w[i:].encode("utf8")
        last = w

def prefix_decode_all(ls):
    """Decompresses a list of strings compressed by prefix_encode().
    """

    last = u''
    for w in ls:
        i = ord(w[0])
        decoded = last[:i] + w[1:].decode("utf8")
        yield decoded
        last = decoded


_nkre = re.compile(r"\D+|\d+", re.UNICODE)
def _nkconv(i):
    try:
        return int(i)
    except ValueError:
        return i.lower()
def natural_key(s):
    """Converts string ``s`` into a tuple that will sort "naturally" (i.e.,
    ``name5`` will come before ``name10`` and ``1`` will come before ``A``).
    This function is designed to be used as the ``key`` argument to sorting
    functions.
    
    :param s: the str/unicode string to convert.
    :rtype: tuple
    """

    # Use _nkre to split the input string into a sequence of
    # digit runs and non-digit runs. Then use _nkconv() to convert
    # the digit runs into ints and the non-digit runs to lowercase.
    return tuple(_nkconv(m) for m in _nkre.findall(s))


class ClosableMixin(object):
    """Mix-in for classes with a close() method to allow them to be used as a
    context manager.
    """

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()


def protected(func):
    """Decorator for storage-access methods. This decorator (a) checks if the
    object has already been closed, and (b) synchronizes on a threading lock.
    The parent object must have 'is_closed' and '_sync_lock' attributes.
    """

    @wraps(func)
    def protected_wrapper(self, *args, **kwargs):
        if self.is_closed:
            raise Exception("%r has been closed" % self)
        if self._sync_lock.acquire(False):
            try:
                return func(self, *args, **kwargs)
            finally:
                self._sync_lock.release()
        else:
            raise Exception("Could not acquire sync lock")

    return protected_wrapper
    

class LRUCache(object):
    def __init__(self, size):
        self.size = size
        self.clock = []
        for i in xrange(0, size):
            self.clock.append([None, False])
        self.hand = 0
        self.data = {}

    def __contains__(self, key):
        return key in self.data
    
    def __getitem__(self, key):
        pos, val = self.data[key]
        self.clock[pos][1] = True
        self.hand = (pos + 1) % self.size
        return val
        
    def __setitem__(self, key, val):
        size = self.size
        hand = self.hand
        clock = self.clock
        data = self.data

        end = (hand or size) - 1
        while True:
            current = clock[hand]
            ref = current[1]
            if ref:
                current[1] = False
                hand = (hand + 1) % size
            elif ref is False or hand == end:
                oldkey = current[0]
                if oldkey in data:
                    del data[oldkey]
                current[0] = key
                current[1] = True
                data[key] = (hand, val)
                hand = (hand + 1) % size
                self.hand = hand
                return






















import unittest

import os, random, time
from shutil import rmtree

from whoosh import fields, index, writing


class TestWriting(unittest.TestCase):
    def make_dir(self, name):
        if not os.path.exists(name):
            os.mkdir(name)
    
    def destroy_dir(self, name):
        try:
            os.rmdir("testindex")
        except:
            raise
    
    def clean_file(self, path):
        if os.path.exists(path):
            os.remove(path)
    
    def test_no_stored(self):
        self.make_dir("testindex")
        schema = fields.Schema(id=fields.ID, text=fields.TEXT)
        ix = index.create_in("testindex", schema, indexname="nostored")
        
        domain = (u"alfa", u"bravo", u"charlie", u"delta", u"echo", u"foxtrot",
                  u"golf", u"hotel", u"india")
        
        w = ix.writer()
        for i in xrange(20):
            w.add_document(id=unicode(i), text=u" ".join(random.sample(domain, 5)))
        w.commit()
        
        r = ix.reader()
        self.assertEqual(sorted([int(id) for id in r.lexicon("id")]), range(20))
        r.close()
    
    def test_asyncwriter(self):
        self.make_dir("testindex")
        schema = fields.Schema(id=fields.ID(stored=True), text=fields.TEXT)
        ix = index.create_in("testindex", schema, indexname="async")
        
        domain = (u"alfa", u"bravo", u"charlie", u"delta", u"echo", u"foxtrot",
                  u"golf", u"hotel", u"india")
        
        writers = []
        # Simulate doing 20 (near-)simultaneous commits. If we weren't using
        # AsyncWriter, at least some of these would fail because the first
        # writer wouldn't be finished yet.
        for i in xrange(20):
            w = writing.AsyncWriter(ix)
            writers.append(w)
            w.add_document(id=unicode(i), text=u" ".join(random.sample(domain, 5)))
            w.commit()
        
        # Wait for all writers to finish before checking the results
        for w in writers:
            if w.running:
                w.join()
        
        # Check whether all documents made it into the index.
        r = ix.reader()
        self.assertEqual(sorted([int(id) for id in r.lexicon("id")]), range(20))
        r.close()
        
        ix.close()
        rmtree("testindex")
        
    def test_asyncwriter_no_stored(self):
        self.make_dir("testindex")
        schema = fields.Schema(id=fields.ID, text=fields.TEXT)
        ix = index.create_in("testindex", schema, indexname="asyncnostored")
        
        domain = (u"alfa", u"bravo", u"charlie", u"delta", u"echo", u"foxtrot",
                  u"golf", u"hotel", u"india")
        
        writers = []
        # Simulate doing 20 (near-)simultaneous commits. If we weren't using
        # AsyncWriter, at least some of these would fail because the first
        # writer wouldn't be finished yet.
        for i in xrange(20):
            w = writing.AsyncWriter(ix)
            writers.append(w)
            w.add_document(id=unicode(i), text=u" ".join(random.sample(domain, 5)))
            w.commit()
        
        # Wait for all writers to finish before checking the results
        for w in writers:
            if w.running:
                w.join()
        
        # Check whether all documents made it into the index.
        r = ix.reader()
        self.assertEqual(sorted([int(id) for id in r.lexicon("id")]), range(20))
        r.close()
        
        ix.close()
        rmtree("testindex")
        
    def test_batchwriter(self):
        self.make_dir("testindex")
        schema = fields.Schema(id=fields.ID, text=fields.TEXT)
        ix = index.create_in("testindex", schema, indexname="batch")
        
        domain = (u"alfa", u"bravo", u"charlie", u"delta", u"echo", u"foxtrot",
                  u"golf", u"hotel", u"india")
        
        w = writing.BatchWriter(ix, period=0.5, limit=10,
                                commitargs={"merge": False})
        for i in xrange(100):
            w.add_document(id=unicode(i), text=u" ".join(random.sample(domain, 5)))
        time.sleep(0.5)
        w.commit(restart=False)
        
        self.assertEqual(len(ix._segments()), 10)
        rmtree("testindex")
        

if __name__ == '__main__':
    unittest.main()



"""
Contains functions implementing the Levenshtein distance algorithm.
"""

def relative(a, b):
    """Returns the relative distance between two strings, in the range
    [0-1] where 1 means total equality.
    """
    d = distance(a,b)
    longer = float(max((len(a), len(b))))
    shorter = float(min((len(a), len(b))))    
    r = ((longer - d) / longer) * (shorter / longer)
    return r

def distance(s, t):
    """Returns the Levenshtein edit distance between two strings."""
    
    m, n = len(s), len(t)
    d = [range(n+1)]
    d += [[i] for i in range(1,m+1)]
    for i in range(0,m):
        for j in range(0,n):
            cost = 1
            if s[i] == t[j]: cost = 0
            d[i+1].append(min(d[i][j+1]+1,  # deletion
                              d[i+1][j]+1,  # insertion
                              d[i][j]+cost) # substitution
                         )
    return d[m][n]




# encoding: utf8
import unittest
import random
from os import mkdir
from os.path import exists
from shutil import rmtree

from whoosh.filedb.filestore import FileStorage, RamStorage
from whoosh.filedb.filetables import (HashReader, HashWriter,
                                      OrderedHashWriter, OrderedHashReader,
                                      StoredFieldWriter, StoredFieldReader,
                                      TermIndexWriter, TermIndexReader)


class TestTables(unittest.TestCase):
    def make_storage(self, dirname):
        if not exists(dirname):
            mkdir(dirname)
        st = FileStorage(dirname)
        return st
    
    def destroy_dir(self, dirname):
        if exists(dirname):
            try:
                rmtree(dirname)
            except:
                pass
    
    def test_termkey(self):
        st = RamStorage()
        tw = TermIndexWriter(st.create_file("test.trm"))
        tw.add(("alfa", u"bravo"), (1.0, 2, 3))
        tw.add((u"alfa", u"Ã¦Ã¯Å�Ãº"), (4.0, 5, 6))
        tw.add((u"text", u"æ—¥æœ¬èªž"), (7.0, 8, 9))
        tw.close()
        
        tr = TermIndexReader(st.open_file("test.trm"))
        self.assertTrue(("alfa", u"bravo") in tr)
        self.assertTrue((u"alfa", u"Ã¦Ã¯Å�Ãº") in tr)
        self.assertTrue((u"text", u"æ—¥æœ¬èªž") in tr)
        
    def test_random_termkeys(self):
        def random_fieldname():
            return "".join(chr(random.randint(65, 90)) for _ in xrange(1, 20))
        
        def random_token():
            return "".join(unichr(random.randint(0, 0xd7ff)) for _ in xrange(1, 20))
        
        domain = sorted([(random_fieldname(), random_token()) for _ in xrange(1000)])
        
        st = RamStorage()
        tw = TermIndexWriter(st.create_file("test.trm"))
        for term in domain:
            tw.add(term, (1.0, 0, 1))
        tw.close()
        
        tr = TermIndexReader(st.open_file("test.trm"))
        for term in domain:
            self.assertTrue(term in tr)
        
    def test_hash(self):
        st = self.make_storage("testindex")
        hwf = st.create_file("test.hsh")
        hw = HashWriter(hwf)
        hw.add("foo", "bar")
        hw.add("glonk", "baz")
        hw.close()
        
        hrf = st.open_file("test.hsh")
        hr = HashReader(hrf)
        self.assertEqual(hr.get("foo"), "bar")
        self.assertEqual(hr.get("baz"), None)
        hr.close()
        
        self.destroy_dir("testindex")
    
    def test_hash_contents(self):
        samp = set((('alfa', 'bravo'), ('charlie', 'delta'), ('echo', 'foxtrot'),
                   ('golf', 'hotel'), ('india', 'juliet'), ('kilo', 'lima'),
                   ('mike', 'november'), ('oskar', 'papa'), ('quebec', 'romeo'),
                   ('sierra', 'tango'), ('ultra', 'victor'), ('whiskey', 'xray')))
        
        st = self.make_storage("testindex")
        hwf = st.create_file("test.hsh")
        hw = HashWriter(hwf)
        hw.add_all(samp)
        hw.close()
        
        hrf = st.open_file("test.hsh")
        hr = HashReader(hrf)
        self.assertEqual(set(hr.items()), samp)
        hr.close()
        
        self.destroy_dir("testindex")
    
    def randstring(self, domain, minlen, maxlen):
        return "".join(random.sample(domain, random.randint(minlen, maxlen)))
    
    def test_random_hash(self):
        domain = "abcdefghijklmnopqrstuvwxyz"
        domain += domain.upper()
        times = 1000
        minlen = 1
        maxlen = len(domain)
        
        samp = dict((self.randstring(domain, minlen, maxlen),
                     self.randstring(domain, minlen, maxlen)) for _ in xrange(times))
        
        st = self.make_storage("testindex")
        hwf = st.create_file("test.hsh")
        hw = HashWriter(hwf)
        for k, v in samp.iteritems():
            hw.add(k, v)
        hw.close()
        
        keys = samp.keys()
        random.shuffle(keys)
        hrf = st.open_file("test.hsh")
        hr = HashReader(hrf)
        for k in keys:
            v = hr[k]
            self.assertEqual(v, samp[k])
        hr.close()
        
        self.destroy_dir("testindex")
    
    def test_ordered_hash(self):
        times = 10000
        st = self.make_storage("testindex")
        hwf = st.create_file("test.hsh")
        hw = HashWriter(hwf)
        hw.add_all(("%08x" % x, str(x)) for x in xrange(times))
        hw.close()
        
        keys = range(times)
        random.shuffle(keys)
        hrf = st.open_file("test.hsh")
        hr = HashReader(hrf)
        for x in keys:
            self.assertEqual(hr["%08x" % x], str(x))
        hr.close()
        
        self.destroy_dir("testindex")
        
    def test_ordered_closest(self):
        keys = ['alfa', 'bravo', 'charlie', 'delta', 'echo', 'foxtrot', 'golf',
                'hotel', 'india', 'juliet', 'kilo', 'lima', 'mike', 'november']
        values = [''] * len(keys)
        
        st = self.make_storage("testindex")
        hwf = st.create_file("test.hsh")
        hw = OrderedHashWriter(hwf)
        hw.add_all(zip(keys, values))
        hw.close()
        
        hrf = st.open_file("test.hsh")
        hr = OrderedHashReader(hrf)
        ck = hr.closest_key
        self.assertEqual(ck(''), 'alfa')
        self.assertEqual(ck(' '), 'alfa')
        self.assertEqual(ck('alfa'), 'alfa')
        self.assertEqual(ck('bravot'), 'charlie')
        self.assertEqual(ck('charlie'), 'charlie')
        self.assertEqual(ck('kiloton'), 'lima')
        self.assertEqual(ck('oskar'), None)
        self.assertEqual(list(hr.keys()), keys)
        self.assertEqual(list(hr.values()), values)
        self.assertEqual(list(hr.keys_from('f')), keys[5:])
        hr.close()
        
        self.destroy_dir("testindex")
        
    def test_stored_fields(self):
        st = self.make_storage("testindex")
        sf = st.create_file("test.sf")
        sfw = StoredFieldWriter(sf, ["a", "b"])
        sfw.append({"a": "hello", "b": "there"})
        sfw.append({"a": "one", "b": "two"})
        sfw.append({"a": "alfa", "b": "bravo"})
        sfw.close()
        
        sf = st.open_file("test.sf")
        sfr = StoredFieldReader(sf)
        self.assertEqual(sfr[0], {"a": "hello", "b": "there"})
        self.assertEqual(sfr[2], {"a": "alfa", "b": "bravo"})
        self.assertEqual(sfr[1], {"a": "one", "b": "two"})
        
        self.destroy_dir("testindex")
        
    def test_termindex(self):
        terms = [("a", "alfa"), ("a", "bravo"), ("a", "charlie"), ("a", "delta"),
                 ("b", "able"), ("b", "baker"), ("b", "dog"), ("b", "easy")]
        st = RamStorage()
        
        tw = TermIndexWriter(st.create_file("test.trm"))
        for i, t in enumerate(terms):
            tw.add(t, (1.0, i * 1000, 1))
        tw.close()
        
        tr = TermIndexReader(st.open_file("test.trm"))
        for i, (t1, t2) in enumerate(zip(tr.keys(), terms)):
            self.assertEqual(t1, t2)
            self.assertEqual(tr.get(t1), (1.0, i * 1000, 1))
        
    

if __name__ == '__main__':
    unittest.main()



import unittest

import os.path, random
from shutil import rmtree

from whoosh import fields, index
from whoosh.filedb.filestore import FileStorage
from whoosh.util import now


class Test(unittest.TestCase):
    def make_index(self, dirname, schema, ixname):
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        st = FileStorage(dirname)
        ix = st.create_index(schema, indexname = ixname)
        return ix
    
    def destroy_index(self, dirname):
        if os.path.exists(dirname):
            try:
                rmtree(dirname)
            except OSError, e:
                raise
    
    def test_20000_small_files(self):
        sc = fields.Schema(id=fields.ID(stored=True), text=fields.TEXT)
        ix = self.make_index("testindex", sc, "ix20000")
        
        domain = ["alfa", "bravo", "charlie", "delta", "echo", "foxtrot",
                  "golf", "hotel", "india", "juliet", "kilo", "lima"]
        
        for i in xrange(20000):
            w = ix.writer()
            w.add_document(id=unicode(i),
                           text = u" ".join(random.sample(domain, 5)))
            w.commit()
        
        ix.optimize()
        #self.destroy_index("testindex")

    def test_20000_batch(self):
        sc = fields.Schema(id=fields.ID(stored=True), text=fields.TEXT)
        ix = self.make_index("testindex", sc, "ix20000")
        
        domain = ["alfa", "bravo", "charlie", "delta", "echo", "foxtrot",
                  "golf", "hotel", "india", "juliet", "kilo", "lima"]
        
        from whoosh.writing import BatchWriter
        w = BatchWriter(ix, limit=100)
        for i in xrange(20000):
            w.add_document(id=unicode(i),
                           text = u" ".join(random.sample(domain, 5)))
        w.commit()
        
        ix.optimize()
        #self.destroy_index("testindex")




if __name__ == "__main__":
    unittest.main()



#===============================================================================
# Copyright 2009 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from collections import defaultdict
from threading import Lock

from whoosh.fields import UnknownFieldError
from whoosh.index import Index
from whoosh.ramdb.ramreading import RamIndexReader
from whoosh.util import protected


class RamIndex(Index):
    def __init__(self, schema):
        self.schema = schema
        self.docnum = 0
        self._sync_lock = Lock()
        self.is_closed = False
        
        self.clear()
    
    def clear(self):
        # Maps fieldname -> a sorted list of term texts in that field
        self.termlists = defaultdict(list)
        
        # Maps fieldnames to dictionaries of term -> posting list
        self.invertedindex = {}
        for fieldname in self.schema.names():
            self.invertedindex[fieldname] = defaultdict(list)
        
        # Maps terms -> index frequencies
        self.indexfreqs = defaultdict(int)
        
        # Maps docnum -> stored field dicts
        self.storedfields = {}
        
        # Maps (docnum, fieldname) -> field length
        self.fieldlengths = defaultdict(int)
        
        # Maps (docnum, fieldname) -> posting list
        self.vectors = {}
        
        # Contains docnums of deleted documents
        self.deleted = set()
        
    def close(self):
        del self.termlists
        del self.invertedindex
        del self.indexfreqs
        del self.storedfields
        del self.fieldlengths
        del self.vectors
        del self.deleted
        self.is_closed = True
    
    def doc_count_all(self):
        return len(self.storedfields)
    
    def doc_count(self):
        return len(self.storedfields) - len(self.deleted)
    
    def field_length(self, fieldname):
        return sum(l for docnum_fieldname, l in self.fieldlengths.iteritems()
                   if docnum_fieldname[1] == fieldname)
        
    def max_field_length(self, fieldname):
        return max(l for docnum_fieldname, l in self.fieldlengths.iteritems()
                   if docnum_fieldname[1] == fieldname)
    
    def reader(self):
        return RamIndexReader(self)
    
    def writer(self):
        return self
    
    @protected
    def add_field(self, *args, **kwargs):
        self.schema.add_field(*args, **kwargs)
    
    @protected
    def remove_field(self, fieldname):
        self.schema.remove_field(fieldname)
        if fieldname in self.termlists:
            del self.termlists[fieldname]
        for fn, text in self.indexfreqs.iterkeys():
            if fn == fieldname:
                del self.indexfreqs[(fn, text)]
        for sfields in self.storedfields.itervalues():
            if fieldname in sfields:
                del sfields[fieldname]
        for docnum, fn in self.fieldlengths.iterkeys():
            if fn == fieldname:
                del self.fieldlengths[(docnum, fn)]
        if fieldname in self.fieldlength_maxes:
            del self.fieldlength_maxes[fieldname]
        for docnum, fn in self.vectors.iterkeys():
            if fn == fieldname:
                del self.vectors[(docnum, fn)]
    
    @protected
    def delete_document(self, docnum, delete=True):
        if delete:
            self.deleted.add(docnum)
        else:
            self.deleted.remove(docnum)
    
    @protected
    def delete_by_term(self, fieldname, text):
        inv = self.invertedindex
        if fieldname in inv:
            terms = inv[fieldname]
            if text in terms:
                postings = terms[text]
                for p in postings:
                    self.deleted.add(p[0])
    
    @protected
    def delete_by_query(self, q, searcher=None):
        s = self.searcher()
        for docnum in q.docs(s):
            self.deleted.add(docnum)
    
    def has_deletions(self):
        return bool(self.deleted)
    
    @protected
    def optimize(self):
        deleted = self.deleted
        
        # Remove deleted documents from stored fields
        storedfields = self.storedfields
        for docnum in deleted:
            del storedfields[docnum]
        
        # Remove deleted documents from inverted index
        removedterms = defaultdict(set)
        for fieldname in self.schema.names():
            inv = self.invertedindex[fieldname]
            for term, postlist in inv.iteritems():
                inv[term] = [x for x in postlist if x[0] not in deleted]
            
            # Remove terms that no longer have any postings after the
            # documents are deleted
            for term in inv.keys():
                if not inv[term]:
                    removedterms[fieldname].add(term)
                    del inv[term]
        
        # If terms were removed as a result of document deletion,
        # update termlists and indexfreqs
        termlists = self.termlists
        for fieldname, removed in removedterms.iteritems():
            termlists[fieldname] = [t for t in termlists[fieldname]
                                   if t not in removed]
            for text in removed:
                del self.indexfreqs[(fieldname, text)]
        
        # Remove documents from field lengths
        fieldlengths = self.fieldlengths
        for docnum, fieldname in fieldlengths.keys():
            if docnum in deleted:
                del fieldlengths[(docnum, fieldname)]
                
        # Remove documents from vectors
        vectors = self.vectors
        for docnum, fieldname in vectors.keys():
            if docnum in deleted:
                del vectors[(docnum, fieldname)]
            
        # Reset deleted list
        self.deleted = set()
        
    @protected
    def add_document(self, **fields):
        schema = self.schema
        invertedindex = self.invertedindex
        indexfreqs = self.indexfreqs
        fieldlengths = self.fieldlengths
        
        fieldnames = [name for name in sorted(fields.keys())
                      if not name.startswith("_")]
        
        storedvalues = {}
        
        for name in fieldnames:
            if name not in schema:
                raise UnknownFieldError("There is no field named %r" % name)
        
        for name in fieldnames:
            value = fields.get(name)
            if value:
                field = schema[name]
                
                newwords = set()
                fielddict = invertedindex[name]
                
                # If the field is indexed, add the words in the value to the
                # index
                if field.indexed:
                    # Count of all terms in the value
                    count = 0
                    # Count of UNIQUE terms in the value
                    unique = 0
                    
                    for w, freq, weight, valuestring in field.index(value):
                        if w not in fielddict:
                            newwords.add(w)
                        fielddict[w].append((self.docnum, weight, valuestring))
                        indexfreqs[(name, w)] += freq
                        count += freq
                        unique += 1
                        
                    self.termlists[name] = sorted(set(self.termlists[name]) | newwords)
                
                    if field.scorable:
                        fieldlengths[(self.docnum, name)] = count
                    
            vector = field.vector
            if vector:
                vlist = sorted((w, weight, valuestring) for w, freq, weight, valuestring
                               in vector.word_values(value))
                self.vectors[(self.docnum, name)] = vlist
            
            if field.stored:
                storedname = "_stored_" + name
                if storedname in fields:
                    stored_value = fields[storedname]
                else :
                    stored_value = value
                
                storedvalues[name] = stored_value
        
        self.storedfields[self.docnum] = storedvalues
        self.docnum += 1
        
    @protected
    def add_reader(self, reader):
        startdoc = self.docnum
        
        has_deletions = reader.has_deletions()
        if has_deletions:
            docmap = {}
        
        fieldnames = set(self.schema.names())
        
        for docnum in xrange(reader.doc_count_all()):
            if (not has_deletions) or (not reader.is_deleted(docnum)):
                d = dict(item for item
                         in reader.stored_fields(docnum).iteritems()
                         if item[0] in fieldnames)
                self.storedfields[self.docnum] = d
                
                if has_deletions:
                    docmap[docnum] = self.docnum
                
                for fieldname, length in reader.doc_field_lengths(docnum):
                    if fieldname in fieldnames:
                        self.fieldlengths[(self.docnum, fieldname)] = length
                        
                for fieldname in reader.vector_names():
                    if (fieldname in fieldnames
                        and reader.has_vector(docnum, fieldname)):
                        vpostreader = reader.vector(docnum, fieldname)
                        self.vectors[(self.docnum, fieldname)] = list(vpostreader.all_items())
                        vpostreader.close()
                
                self.docnum += 1
                
        for fieldname, text, _, _ in reader:
            if fieldname in fieldnames:
                postreader = reader.postings(fieldname, text)
                while postreader.is_active():
                    docnum = postreader.id()
                    valuestring = postreader.value()
                    weight = postreader.weight()
                    if has_deletions:
                        newdoc = docmap[docnum]
                    else:
                        newdoc = startdoc + docnum
                    self.invertedindex[fieldname][text].append((newdoc,
                                                                weight,
                                                                valuestring))
                    postreader.next()
    
    
    





    


import unittest
from os import mkdir
from os.path import exists
from shutil import rmtree

from whoosh import analysis, fields, formats, index, qparser
from whoosh.filedb.filestore import FileStorage
from whoosh.searching import Searcher


class TestVectors(unittest.TestCase):
    def make_index(self, dirname, schema, indexname):
        if not exists(dirname):
            mkdir(dirname)
        st = FileStorage(dirname)
        ix = st.create_index(schema, indexname=indexname)
        return ix
    
    def destroy_index(self, dirname):
        if exists(dirname):
            rmtree(dirname)
    
    def test_vector_reading(self):
        a = analysis.StandardAnalyzer()
        schema = fields.Schema(title = fields.TEXT,
                               content = fields.TEXT(vector=formats.Frequency(analyzer=a)))
        ix = self.make_index("testindex", schema, "vector_reading")
        try:
            writer = ix.writer()
            writer.add_document(title=u"one",
                                content=u"This is the story of the black hole story")
            writer.commit()
            
            reader = ix.reader()
            self.assertEqual(list(reader.vector_as("frequency", 0, "content")),
                             [(u'black', 1), (u'hole', 1), (u'story', 2)])
        finally:
            pass
            #self.destroy_index("testindex")
    
    def test_vector_merge(self):
        a = analysis.StandardAnalyzer()
        schema = fields.Schema(title = fields.TEXT,
                               content = fields.TEXT(vector=formats.Frequency(analyzer=a)))
        ix = self.make_index("testindex", schema, "vector_merge")
        try:
            writer = ix.writer()
            writer.add_document(title=u"one",
                                content=u"This is the story of the black hole story")
            writer.commit()
            
            writer = ix.writer()
            writer.add_document(title=u"two",
                                content=u"You can read along in your book")
            writer.commit()
            
            searcher = Searcher(ix)
            reader = searcher.reader()
            
            docnum = searcher.document_number(title=u"one")
            vec = list(reader.vector_as("frequency", docnum, "content"))
            self.assertEqual(vec, [(u'black', 1), (u'hole', 1), (u'story', 2)])
            
            docnum = searcher.document_number(title=u"two")
            vec = list(reader.vector_as("frequency", docnum, "content"))
            self.assertEqual(vec, [(u'along', 1), (u'book', 1), (u'read', 1)])
        finally:
            pass
            #self.destroy_index("testindex")
            
    def test_vector_unicode(self):
        a = analysis.StandardAnalyzer()
        schema = fields.Schema(content = fields.TEXT(vector=formats.Frequency(analyzer=a)))
        ix = self.make_index("testindex", schema, "vector_unicode")
        try:
            writer = ix.writer()
            writer.add_document(content=u"\u1234\u2345\u3456 \u4567\u5678\u6789")
            writer.add_document(content=u"\u0123\u1234\u4567 \u4567\u5678\u6789")
            writer.commit()
            
            writer = ix.writer()
            writer.add_document(content=u"\u2345\u3456\u4567 \u789a\u789b\u789c")
            writer.add_document(content=u"\u0123\u1234\u4567 \u2345\u3456\u4567")
            writer.commit()
            
            reader = ix.reader()
            vec = list(reader.vector_as("frequency", 0, "content"))
            self.assertEqual(vec, [(u'\u3456\u4567', 1), (u'\u789a\u789b\u789c', 1)])
        finally:
            pass
            #self.destroy_index("testindex")


if __name__ == '__main__':
    unittest.main()



#===============================================================================
# Copyright 2009 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import mmap, os, sys
from array import array
from copy import copy
from cPickle import dump as dump_pickle
from cPickle import load as load_pickle
from struct import calcsize

from whoosh.system import (_INT_SIZE, _SHORT_SIZE, _FLOAT_SIZE, _LONG_SIZE,
                           pack_sbyte, pack_ushort, pack_int, pack_uint,
                           pack_long, pack_float,
                           unpack_sbyte, unpack_ushort, unpack_int,
                           unpack_uint, unpack_long, unpack_float, IS_LITTLE)
from whoosh.util import (varint, read_varint, signed_varint,
                         decode_signed_varint, float_to_byte, byte_to_float)


_SIZEMAP = dict((typecode, calcsize(typecode)) for typecode in "bBiIhHqQf")
_ORDERMAP = {"little": "<", "big": ">"}

_types = (("sbyte", "b"), ("ushort", "H"), ("int", "i"),
          ("long", "q"), ("float", "f"))


# Main function

class StructFile(object):
    """Returns a "structured file" object that wraps the given file object and
    provides numerous additional methods for writing structured data, such as
    "write_varint" and "write_long".
    """

    def __init__(self, fileobj, name=None, onclose=None, mapped=True):
        self.file = fileobj
        self._name = name
        self.onclose = onclose
        self.is_closed = False

        for attr in ("read", "write", "tell", "seek", "truncate"):
            if hasattr(fileobj, attr):
                setattr(self, attr, getattr(fileobj, attr))

        # If mapped is True, set the 'map' attribute to a memory-mapped
        # representation of the file. Otherwise, the fake 'map' that set up by
        # the base class will be used.
        if mapped and hasattr(fileobj, "mode") and "r" in fileobj.mode:
            fd = fileobj.fileno()
            self.size = os.fstat(fd).st_size
            try:
                self.map = mmap.mmap(fd, self.size, access=mmap.ACCESS_READ)
            except OSError:
                self._setup_fake_map()
        else:
            self._setup_fake_map()
            
        self.is_real = hasattr(fileobj, "fileno")

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self._name)

    def flush(self):
        """Flushes the buffer of the wrapped file. This is a no-op if the
        wrapped file does not have a flush method.
        """
        if hasattr(self.file, "flush"):
            self.file.flush()

    def close(self):
        """Closes the wrapped file. This is a no-op if the wrapped file does
        not have a close method.
        """

        if self.is_closed:
            raise Exception("This file is already closed")
        del self.map
        if self.onclose:
            self.onclose(self)
        if hasattr(self.file, "close"):
            self.file.close()
        self.is_closed = True

    def _setup_fake_map(self):
        _self = self
        class fakemap(object):
            def __getitem__(self, slice):
                if isinstance(slice, (int, long)):
                    _self.seek(slice)
                    return _self.read(1)
                else:
                    _self.seek(slice.start)
                    return _self.read(slice.stop - slice.start)
        self.map = fakemap()

    def write_string(self, s):
        """Writes a string to the wrapped file. This method writes the length
        of the string first, so you can read the string back without having to
        know how long it was.
        """
        self.write_varint(len(s))
        self.file.write(s)

    def write_string2(self, s):
        self.write(pack_ushort(len(s)) + s)

    def read_string(self):
        """Reads a string from the wrapped file.
        """
        return self.file.read(self.read_varint())

    def read_string2(self):
        l = self.read_ushort()
        return self.read(l)

    def skip_string(self):
        l = self.read_varint()
        self.seek(l, 1)

    def write_varint(self, i):
        """Writes a variable-length unsigned integer to the wrapped file.
        """
        self.file.write(varint(i))

    def write_svarint(self, i):
        """Writes a variable-length signed integer to the wrapped file.
        """
        self.file.write(signed_varint(i))

    def read_varint(self):
        """Reads a variable-length encoded unsigned integer from the wrapped file.
        """
        return read_varint(self.file.read)

    def read_svarint(self):
        """Reads a variable-length encoded signed integer from the wrapped file.
        """
        return decode_signed_varint(read_varint(self.file.read))

    def write_byte(self, n):
        """Writes a single byte to the wrapped file, shortcut for
        ``file.write(chr(n))``.
        """
        self.file.write(chr(n))

    def read_byte(self):
        return ord(self.file.read(1))

    def get_byte(self, position):
        return ord(self.map[position])

    def write_8bitfloat(self, f, mantissabits=5, zeroexp=2):
        """Writes a byte-sized representation of floating point value f to the
        wrapped file.
        
        :param mantissabits: the number of bits to use for the mantissa
            (with the rest used for the exponent).
        :param zeroexp: the zero point for the exponent.
        """

        self.write_byte(float_to_byte(f, mantissabits, zeroexp))

    def read_8bitfloat(self, mantissabits=5, zeroexp=2):
        """Reads a byte-sized representation of a floating point value.
        
        :param mantissabits: the number of bits to use for the mantissa
            (with the rest used for the exponent).
        :param zeroexp: the zero point for the exponent.
        """
        return byte_to_float(self.read_byte(), mantissabits, zeroexp)

    def write_pickle(self, obj, protocol= -1):
        """Writes a pickled representation of obj to the wrapped file.
        """
        dump_pickle(obj, self.file, protocol)

    def read_pickle(self):
        """Reads a pickled object from the wrapped file.
        """
        return load_pickle(self.file)

    def write_sbyte(self, n):
        self.file.write(pack_sbyte(n))
    def write_int(self, n):
        self.file.write(pack_int(n))
    def write_uint(self, n):
        self.file.write(pack_uint(n))
    def write_ushort(self, n):
        self.file.write(pack_ushort(n))
    def write_long(self, n):
        self.file.write(pack_long(n))
    def write_float(self, n):
        self.file.write(pack_float(n))
    def write_array(self, arry):
        if IS_LITTLE:
            arry = copy(arry)
            arry.byteswap()
        if self.is_real:
            arry.tofile(self.file)
        else:
            self.file.write(arry.tostring())

    def read_sbyte(self):
        return unpack_sbyte(self.file.read(1))[0]
    def read_int(self):
        return unpack_int(self.file.read(_INT_SIZE))[0]
    def read_uint(self):
        return unpack_uint(self.file.read(_INT_SIZE))[0]
    def read_ushort(self):
        return unpack_ushort(self.file.read(_SHORT_SIZE))[0]
    def read_long(self):
        return unpack_long(self.file.read(_LONG_SIZE))[0]
    def read_float(self):
        return unpack_float(self.file.read(_FLOAT_SIZE))[0]
    def read_array(self, typecode, length):
        a = array(typecode)
        if self.is_real:
            a.fromfile(self.file, length)
        else:
            a.fromstring(self.file.read(length * _SIZEMAP[typecode]))
        if IS_LITTLE: a.byteswap()
        return a

    def get_sbyte(self, position):
        return unpack_sbyte(self.map[position:position + 1])[0]
    def get_int(self, position):
        return unpack_int(self.map[position:position + _INT_SIZE])[0]
    def get_uint(self, position):
        return unpack_uint(self.map[position:position + _INT_SIZE])[0]
    def get_ushort(self, position):
        return unpack_ushort(self.map[position:position + _SHORT_SIZE])[0]
    def get_long(self, position):
        return unpack_long(self.map[position:position + _LONG_SIZE])[0]
    def get_float(self, position):
        return unpack_float(self.map[position:position + _FLOAT_SIZE])[0]
    def get_array(self, position, typecode, length):
        source = self.map[position:position + length * _SIZEMAP[typecode]]
        a = array(typecode)
        a.fromstring(source)
        if IS_LITTLE: a.byteswap()
        return a
















# coding=utf-8

import unittest

from whoosh import analysis, fields, qparser, query


class TestQueryParser(unittest.TestCase):
    def test_fields(self):
        s = fields.Schema(content=fields.TEXT, title=fields.TEXT, id=fields.ID)
        
        qp = qparser.QueryParser("content", s)
        q = qp.parse(u"test")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.fieldname, "content")
        self.assertEqual(q.text, "test")
        
        mq = qparser.MultifieldParser(("title", "content"), s)
        q = mq.parse(u"test")
        self.assertEqual(q.__class__, query.Or)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[0].fieldname, "title")
        self.assertEqual(q[1].fieldname, "content")
        self.assertEqual(q[0].text, "test")
        self.assertEqual(q[1].text, "test")
        
        q = mq.parse(u"title:test")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.fieldname, "title")
        self.assertEqual(q.text, "test")
    
    def test_colonspace(self):
        s = fields.Schema(content=fields.TEXT, url=fields.ID)
        
        qp = qparser.QueryParser("content", s)
        q = qp.parse(u"url:test")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.fieldname, "url")
        self.assertEqual(q.text, "test")
        
        q = qp.parse(u"url: test")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[0].fieldname, "content")
        self.assertEqual(q[1].fieldname, "content")
        self.assertEqual(q[0].text, "url")
        self.assertEqual(q[1].text, "test")
        
        q = qp.parse(u"url:")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.fieldname, "content")
        self.assertEqual(q.text, "url")
    
    def test_andor(self):
        qp  = qparser.QueryParser("a")
        q = qp.parse("a AND b OR c AND d OR e AND f")
        self.assertEqual(unicode(q), u"(a:a AND (a:b OR (a:c AND (a:d OR (a:e AND a:f)))))")
    
    def test_andnot(self):
        qp = qparser.QueryParser("content")
        q = qp.parse(u"this ANDNOT that")
        self.assertEqual(q.__class__, query.AndNot)
        self.assertEqual(q.positive.__class__, query.Term)
        self.assertEqual(q.negative.__class__, query.Term)
        self.assertEqual(q.positive.text, "this")
        self.assertEqual(q.negative.text, "that")
        
        q = qp.parse(u"foo ANDNOT bar baz")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 2)
        self.assertEqual(q[0].__class__, query.AndNot)
        self.assertEqual(q[1].__class__, query.Term)
        
        q = qp.parse(u"foo fie ANDNOT bar baz")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 3)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[1].__class__, query.AndNot)
        self.assertEqual(q[2].__class__, query.Term)
        
        q = qp.parse(u"a AND b ANDNOT c")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 2)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[0].text, "a")
        self.assertEqual(q[1].__class__, query.AndNot)
        self.assertEqual(q[1].positive.text, "b")
        self.assertEqual(q[1].negative.text, "c")
    
    def test_boost(self):
        qp = qparser.QueryParser("content")
        q = qp.parse(u"this^3 fn:that^0.5 5.67 hi^5x")
        self.assertEqual(q[0].boost, 3.0)
        self.assertEqual(q[1].boost, 0.5)
        self.assertEqual(q[1].fieldname, "fn")
        self.assertEqual(q[2].text, "5.67")
        self.assertEqual(q[3].text, "hi^5x")
        
        q = qp.parse("alfa (bravo OR charlie)^2.5 ^3")
        self.assertEqual(len(q), 3)
        self.assertEqual(q[0].boost, 1.0)
        self.assertEqual(q[1].boost, 2.5)
        self.assertEqual(q[2].text, "^3")
        
    def test_wildcard1(self):
        qp = qparser.QueryParser("content")
        q = qp.parse(u"hello *the?e* ?star*s? test")
        self.assertEqual(len(q), 4)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[0].text, "hello")
        self.assertEqual(q[1].__class__, query.Wildcard)
        self.assertEqual(q[1].text, "*the?e*")
        self.assertEqual(q[2].__class__, query.Wildcard)
        self.assertEqual(q[2].text, "?star*s?")
        self.assertEqual(q[3].__class__, query.Term)
        self.assertEqual(q[3].text, "test")
        
    def test_wildcard2(self):
        qp = qparser.QueryParser("content")
        q = qp.parse(u"*the?e*")
        self.assertEqual(q.__class__, query.Wildcard)
        self.assertEqual(q.text, "*the?e*")
        
    def test_parse_fieldname_underscores(self):
        s = fields.Schema(my_name=fields.ID(stored=True), my_value=fields.TEXT)
        qp = qparser.QueryParser("my_value", schema=s)
        q = qp.parse(u"my_name:Green")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.fieldname, "my_name")
        self.assertEqual(q.text, "Green")
    
    def test_endstar(self):
        qp = qparser.QueryParser("text")
        q = qp.parse(u"word*")
        self.assertEqual(q.__class__, query.Prefix)
        self.assertEqual(q.text, "word")
        
        q = qp.parse(u"first* second")
        self.assertEqual(q[0].__class__, query.Prefix)
        self.assertEqual(q[0].text, "first")
    
    def test_singlequotes(self):
        qp = qparser.QueryParser("text")
        q = qp.parse("hell's hot 'i stab at thee'")
        self.assertEqual(q.__class__.__name__, 'And')
        self.assertEqual(len(q), 3)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[2].__class__, query.Term)
        self.assertEqual(q[0].text, "hell's")
        self.assertEqual(q[1].text, "hot")
        self.assertEqual(q[2].text, "i stab at thee")
        
        q = qp.parse("alfa zulu:'bravo charlie' delta")
        self.assertEqual(q.__class__.__name__, 'And')
        self.assertEqual(len(q), 3)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[2].__class__, query.Term)
        self.assertEqual((q[0].fieldname, q[0].text), ("text", "alfa"))
        self.assertEqual((q[1].fieldname, q[1].text), ("zulu", "bravo charlie"))
        self.assertEqual((q[2].fieldname, q[2].text), ("text", "delta"))
    
#    def test_escaping(self):
#        qp = qparser.QueryParser("text")
#        
#        q = qp.parse(r'big\small')
#        self.assertEqual(q.__class__, query.Term, q)
#        self.assertEqual(q.text, "bigsmall")
#        
#        q = qp.parse(r'big\\small')
#        self.assertEqual(q.__class__, query.Term)
#        self.assertEqual(q.text, r'big\small')
#        
#        q = qp.parse(r'http\:example')
#        self.assertEqual(q.__class__, query.Term)
#        self.assertEqual(q.fieldname, "text")
#        self.assertEqual(q.text, "http:example")
#        
#        q = qp.parse(r'hello\ there')
#        self.assertEqual(q.__class__, query.Term)
#        self.assertEqual(q.text, "hello there")
#        
#        q = qp.parse(r'\[start\ TO\ end\]')
#        self.assertEqual(q.__class__, query.Term)
#        self.assertEqual(q.text, "[start TO end]")
#    
#        schema = fields.Schema(text=fields.TEXT)
#        qp = qparser.QueryParser("text")
#        q = qp.parse(r"http\:\/\/www\.example\.com")
#        self.assertEqual(q.__class__, query.Term)
#        self.assertEqual(q.text, "http://www.example.com")
#        
#        q = qp.parse(u"\\\\")
#        self.assertEqual(q.__class__, query.Term)
#        self.assertEqual(q.text, "\\")
    
#    def test_escaping_wildcards(self):
#        qp = qparser.QueryParser("text")
#        
#        q = qp.parse(u"a*b*c?d")
#        self.assertEqual(q.__class__, query.Wildcard)
#        self.assertEqual(q.text, "a*b*c?d")
#        
#        q = qp.parse(u"a*b\\*c?d")
#        self.assertEqual(q.__class__, query.Wildcard)
#        self.assertEqual(q.text, "a*b*c?d")
#        
#        q = qp.parse(u"a*b\\\\*c?d")
#        self.assertEqual(q.__class__, query.Wildcard)
#        self.assertEqual(q.text, u'a*b\\*c?d')
#        
#        q = qp.parse(u"ab*")
#        self.assertEqual(q.__class__, query.Prefix)
#        self.assertEqual(q.text, u"ab")
#        
#        q = qp.parse(u"ab\\\\*")
#        self.assertEqual(q.__class__, query.Wildcard)
#        self.assertEqual(q.text, u"ab\\*")
        
    def test_phrase(self):
        qp = qparser.QueryParser("content")
        q = qp.parse('"alfa bravo" "charlie delta echo"^2.2 test:"foxtrot golf"')
        self.assertEqual(q[0].__class__, query.Phrase)
        self.assertEqual(q[0].words, ["alfa", "bravo"])
        self.assertEqual(q[1].__class__, query.Phrase)
        self.assertEqual(q[1].words, ["charlie", "delta", "echo"])
        self.assertEqual(q[1].boost, 2.2)
        self.assertEqual(q[2].__class__, query.Phrase)
        self.assertEqual(q[2].words, ["foxtrot", "golf"])
        self.assertEqual(q[2].fieldname, "test")
        
    def test_weird_characters(self):
        qp = qparser.QueryParser("content")
        q = qp.parse(u".abcd@gmail.com")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.text, u".abcd@gmail.com")
        q = qp.parse(u"r*")
        self.assertEqual(q.__class__, query.Prefix)
        self.assertEqual(q.text, u"r")
        q = qp.parse(u".")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.text, u".")
        q = qp.parse(u"?")
        self.assertEqual(q.__class__, query.Wildcard)
        self.assertEqual(q.text, u"?")
        
    def test_euro_chars(self):
        schema = fields.Schema(text=fields.TEXT)
        qp = qparser.QueryParser("text", schema=schema)
        q = qp.parse(u"straße")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.text, u"straße")
    
    def test_star(self):
        schema = fields.Schema(text = fields.TEXT(stored=True))
        qp = qparser.QueryParser("text", schema=schema)
        q = qp.parse(u"*")
        self.assertEqual(q.__class__, query.Every)
        
        q = qp.parse(u"*h?ll*")
        self.assertEqual(q.__class__, query.Wildcard)
        self.assertEqual(q.text, u"*h?ll*")
        
        q = qp.parse(u"h?pe")
        self.assertEqual(q.__class__, query.Wildcard)
        self.assertEqual(q.text, u"h?pe")
        
        q = qp.parse(u"*? blah")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(q[0].__class__, query.Wildcard)
        self.assertEqual(q[0].text, u"*?")
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[1].text, u"blah")
        
        q = qp.parse(u"*ending")
        self.assertEqual(q.__class__, query.Wildcard)
        self.assertEqual(q.text, u"*ending")
        
        q = qp.parse(u"*q")
        self.assertEqual(q.__class__, query.Wildcard)
        self.assertEqual(q.text, u"*q")

    def test_range(self):
        schema = fields.Schema(name=fields.ID(stored=True), text = fields.TEXT(stored=True))
        qp = qparser.QueryParser("text", schema=schema)
        
        q = qp.parse(u"[alfa to bravo}")
        self.assertEqual(q.__class__, query.TermRange)
        self.assertEqual(q.start, "alfa")
        self.assertEqual(q.end, "bravo")
        self.assertEqual(q.startexcl, False)
        self.assertEqual(q.endexcl, True)
        
        q = qp.parse(u"['hello there' to 'what ever']")
        self.assertEqual(q.__class__, query.TermRange)
        self.assertEqual(q.start, "hello there")
        self.assertEqual(q.end, "what ever")
        self.assertEqual(q.startexcl, False)
        self.assertEqual(q.endexcl, False)
        
        q = qp.parse(u"name:{'to' to 'b'}")
        self.assertEqual(q.__class__, query.TermRange)
        self.assertEqual(q.start, "to")
        self.assertEqual(q.end, "b")
        self.assertEqual(q.startexcl, True)
        self.assertEqual(q.endexcl, True)
        
        q = qp.parse(u"name:{'a' to 'to']")
        self.assertEqual(q.__class__, query.TermRange)
        self.assertEqual(q.start, "a")
        self.assertEqual(q.end, "to")
        self.assertEqual(q.startexcl, True)
        self.assertEqual(q.endexcl, False)
        
        q = qp.parse(u"name:[a to to]")
        self.assertEqual(q.__class__, query.TermRange)
        self.assertEqual(q.start, "a")
        self.assertEqual(q.end, "to")
        
        q = qp.parse(u"name:[to to b]")
        self.assertEqual(q.__class__, query.TermRange)
        self.assertEqual(q.start, "to")
        self.assertEqual(q.end, "b")
        
        q = qp.parse(u"[alfa to alfa]")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.text, "alfa")
        
        q = qp.parse(u"Ind* AND name:[d TO]")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(q[0].__class__, query.Prefix)
        self.assertEqual(q[1].__class__, query.TermRange)
        self.assertEqual(q[0].text, "ind")
        self.assertEqual(q[1].start, "d")
        self.assertEqual(q[1].fieldname, "name")
        
        q = qp.parse(u"name:[d TO]")
        self.assertEqual(q.__class__, query.TermRange)
        self.assertEqual(q.start, "d")
        self.assertEqual(q.fieldname, "name")
    
    def test_numeric_range(self):
        schema = fields.Schema(id=fields.STORED, number=fields.NUMERIC)
        qp = qparser.QueryParser("number", schema=schema)
        
        teststart = 40
        testend = 100
        
        q = qp.parse("[%s to *]" % teststart)
        self.assertEqual(q, query.NullQuery)
        
        q = qp.parse("[%s to]" % teststart)
        self.assertEqual(q.__class__, query.NumericRange)
        self.assertEqual(q.start, teststart)
        self.assertEqual(q.end, None)
        
        q = qp.parse("[to %s]" % testend)
        self.assertEqual(q.__class__, query.NumericRange)
        self.assertEqual(q.start, None)
        self.assertEqual(q.end, testend)
        
        q = qp.parse("[%s to %s]" % (teststart, testend))
        self.assertEqual(q.__class__, query.NumericRange)
        self.assertEqual(q.start, teststart)
        self.assertEqual(q.end, testend)
        
    def test_regressions(self):
        qp = qparser.QueryParser("f")
        
        # From 0.3.18, these used to require escaping. Mostly good for
        # regression testing.
        self.assertEqual(qp.parse(u"re-inker"), query.Term("f", u"re-inker"))
        self.assertEqual(qp.parse(u"0.7 wire"),
                         query.And([query.Term("f", u"0.7"), query.Term("f", u"wire")]))
        self.assertEqual(qp.parse(u"daler-rowney pearl 'bell bronze'"),
                         query.And([query.Term("f", u"daler-rowney"),
                                    query.Term("f", u"pearl"),
                                    query.Term("f", u"bell bronze")]))
        self.assertEqual(qp.parse(u'22" BX'),
                         query.And([query.Term("f", u'22"'), query.Term("f", "BX")]))
        
    def test_empty_ranges(self):
        schema = fields.Schema(name=fields.TEXT, num=fields.NUMERIC,
                               date=fields.DATETIME)
        qp = qparser.QueryParser("text", schema=schema)
        
        for fname in ("name", "date"):
            q = qp.parse("%s:[to]" % fname)
            self.assertEqual(q.__class__, query.Every)
    
    def test_empty_numeric_range(self):
        schema = fields.Schema(id=fields.ID, num=fields.NUMERIC)
        qp = qparser.QueryParser("num", schema=schema)
        q = qp.parse("num:[to]")
        self.assertEqual(q.__class__, query.NumericRange)
        self.assertEqual(q.start, None)
        self.assertEqual(q.end, None)
    
    def test_stopped(self):
        schema = fields.Schema(text = fields.TEXT)
        qp = qparser.QueryParser("text", schema=schema)
        q = qp.parse(u"a b")
        self.assertEqual(q, query.NullQuery)
        
    def test_analyzing_terms(self):
        schema = fields.Schema(text=fields.TEXT(analyzer=analysis.StemmingAnalyzer()))
        qp = qparser.QueryParser("text", schema=schema)
        q = qp.parse(u"Indexed!")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.text, "index")
        
    def test_simple(self):
        parser = qparser.SimpleParser("x")
        q = parser.parse(u"alfa bravo charlie delta")
        self.assertEqual(unicode(q), u"(x:alfa OR x:bravo OR x:charlie OR x:delta)")
        
        q = parser.parse(u"alfa +bravo charlie delta")
        self.assertEqual(unicode(q), u"(x:bravo ANDMAYBE (x:alfa OR x:charlie OR x:delta))")
        
        q = parser.parse(u"alfa +bravo -charlie delta")
        self.assertEqual(unicode(q), u"(x:bravo ANDMAYBE (x:alfa OR x:delta)) ANDNOT x:charlie")
        
        q = parser.parse(u"- alfa +bravo + delta")
        self.assertEqual(unicode(q), u"(x:bravo AND x:delta) ANDNOT x:alfa")
    
    def test_dismax(self):
        parser = qparser.DisMaxParser({"body": 0.8, "title": 2.5})
        q = parser.parse(u"alfa bravo charlie")
        self.assertEqual(unicode(q), u"(DisMax(body:alfa^0.8 title:alfa^2.5) OR DisMax(body:bravo^0.8 title:bravo^2.5) OR DisMax(body:charlie^0.8 title:charlie^2.5))")
        
        q = parser.parse(u"alfa +bravo charlie")
        self.assertEqual(unicode(q), u"(DisMax(body:bravo^0.8 title:bravo^2.5) ANDMAYBE (DisMax(body:alfa^0.8 title:alfa^2.5) OR DisMax(body:charlie^0.8 title:charlie^2.5)))")
        
        q = parser.parse(u"alfa -bravo charlie")
        self.assertEqual(unicode(q), u"(DisMax(body:alfa^0.8 title:alfa^2.5) OR DisMax(body:charlie^0.8 title:charlie^2.5)) ANDNOT DisMax(body:bravo^0.8 title:bravo^2.5)")
        
        q = parser.parse(u"alfa -bravo +charlie")
        self.assertEqual(unicode(q), u"(DisMax(body:charlie^0.8 title:charlie^2.5) ANDMAYBE DisMax(body:alfa^0.8 title:alfa^2.5)) ANDNOT DisMax(body:bravo^0.8 title:bravo^2.5)")
    
    def test_many_clauses(self):
        parser = qparser.QueryParser("content")
        qs = "1" + (" OR 1" * 1000)
        parser.parse(qs)
        
        parser = qparser.QueryParser("content")
        parser.parse(qs)
        
    def test_roundtrip(self):
        parser = qparser.QueryParser("a")
        q = parser.parse(u"a OR ((b AND c AND d AND e) OR f OR g) ANDNOT h")
        self.assertEqual(unicode(q), u"(a:a OR ((a:b AND a:c AND a:d AND a:e) OR a:f OR a:g) ANDNOT a:h)")
        
    def test_math(self):
        tk = analysis.RegexTokenizer(r"([A-Za-z_\\]+)|(-?[0-9.]+)|([^A-Za-z_0-9. \t\r\n]+)")
        ana = tk | analysis.LowercaseFilter()
        print [t.text for t in ana(u"x2+6x+7")]
        
        schema = fields.Schema(a=fields.TEXT(analyzer=ana))
        parser = qparser.QueryParser("a", schema=schema)
        print parser.parse(u"+")
        print parser.parse(u"5+4")

        


if __name__ == '__main__':
    unittest.main()



import unittest
import inspect
from random import choice, randint

from whoosh import query, scoring
from whoosh.fields import *
from whoosh.filedb.filestore import RamStorage
from whoosh.searching import Searcher
from whoosh.util import permutations


class TestWeightings(unittest.TestCase):
    def _weighting_classes(self, ignore):
        # Get all the subclasses of Weighting in whoosh.scoring
        return [c for name, c in inspect.getmembers(scoring, inspect.isclass)
                if scoring.Weighting in c.__bases__ and c not in ignore]
        
    def test_all(self):
        domain = [u"alfa", u"bravo", u"charlie", u"delta", u"echo", u"foxtrot"]
        schema = Schema(text=TEXT)
        storage = RamStorage()
        ix = storage.create_index(schema)
        w = ix.writer()
        for _ in xrange(100):
            w.add_document(text=u" ".join(choice(domain) for i in xrange(randint(10, 20))))
        w.commit()
        
        # List ABCs that should not be tested
        abcs = ()
        # provide initializer arguments for any weighting classes that require them
        init_args = {"MultiWeighting": ([scoring.BM25F()], {"text": scoring.Frequency()}),
                     "ReverseWeighting": ([scoring.BM25F()], {})}
        
        for wclass in self._weighting_classes(abcs):
            try:
                if wclass.__name__ in init_args:
                    args, kwargs = init_args[wclass.__name__]
                    weighting = wclass(*args, **kwargs)
                else:
                    weighting = wclass()
            except TypeError, e:
                raise TypeError("Error instantiating %r: %s" % (wclass, e))
            searcher = ix.searcher(weighting=weighting)
            
            try:
                for word in domain:
                    searcher.search(query.Term("text", word))
            except Exception, e:
                e.msg = "Error searching with %r: %s" % (wclass, e)
                raise
    
    def test_compatibility(self):
        from whoosh.scoring import Weighting
        
        # This is the old way of doing a custom weighting model, check that
        # it's still supported...
        class LegacyWeighting(Weighting):
            use_final = True
            
            def score(self, searcher, fieldname, text, docnum, weight):
                return weight + 0.5
            
            def final(self, searcher, docnum, score):
                return score * 1.5
        
        schema = Schema(text=TEXT)
        ix = RamStorage().create_index(schema)
        w = ix.writer()
        domain = "alfa bravo charlie delta".split()
        for ls in permutations(domain, 3):
            w.add_document(text=u" ".join(ls))
        w.commit()
        
        s = ix.searcher(weighting=LegacyWeighting())
        r = s.search(query.Term("text", u"bravo"))
        self.assertEqual(r.score(0), 2.25)
        






#===============================================================================
# Copyright 2010 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import os
from multiprocessing import Process, Queue

from whoosh.filedb.filetables import LengthWriter, LengthReader
from whoosh.filedb.filewriting import SegmentWriter
from whoosh.filedb.pools import (imerge, PoolBase, read_run, TempfilePool,
                                 write_postings)
from whoosh.filedb.structfile import StructFile
from whoosh.writing import IndexWriter
from whoosh.util import now


# Multiprocessing writer

class SegmentWritingTask(Process):
    def __init__(self, storage, indexname, segmentname, kwargs, postingqueue):
        Process.__init__(self)
        self.storage = storage
        self.indexname = indexname
        self.segmentname = segmentname
        self.kwargs = kwargs
        self.postingqueue = postingqueue
        
        self.segment = None
        self.running = True
    
    def run(self):
        pqueue = self.postingqueue
        
        index = self.storage.open_index(self.indexname)
        writer = SegmentWriter(index, name=self.segmentname, lock=False, **self.kwargs)
        
        while self.running:
            args = pqueue.get()
            if args is None:
                break
            
            writer.add_document(**args)
        
        if not self.running:
            writer.cancel()
            self.terminate()
        else:
            writer.pool.finish(writer.docnum, writer.lengthfile,
                               writer.termsindex, writer.postwriter)
            self._segment = writer._getsegment()
    
    def get_segment(self):
        return self._segment
    
    def cancel(self):
        self.running = False


class MultiSegmentWriter(IndexWriter):
    def __init__(self, index, procs=2, **writerargs):
        self.index = index
        self.lock = index.storage.lock(index.indexname + "_LOCK")
        self.tasks = []
        self.postingqueue = Queue()
        #self.resultqueue = Queue()
        
        names = [index._next_segment_name() for _ in xrange(procs)]
        
        self.tasks = [SegmentWritingTask(index.storage, index.indexname,
                                         segname, writerargs, self.postingqueue)
                      for segname in names]
        for task in self.tasks:
            task.start()
        
    def add_document(self, **args):
        self.postingqueue.put(args)
        
    def cancel(self):
        for task in self.tasks:
            task.cancel()
        self.lock.release()
        
    def commit(self):
        procs = len(self.tasks)
        for _ in xrange(procs):
            self.postingqueue.put(None)
        for task in self.tasks:
            print "Joining", task
            task.join()
            self.index.segments.append(task.get_segment())
        self.index.commit()
        self.lock.release()


# Multiprocessing pool

class PoolWritingTask(Process):
    def __init__(self, schema, dir, postingqueue, resultqueue, limitmb):
        Process.__init__(self)
        self.schema = schema
        self.dir = dir
        self.postingqueue = postingqueue
        self.resultqueue = resultqueue
        self.limitmb = limitmb
        
    def run(self):
        pqueue = self.postingqueue
        rqueue = self.resultqueue
        
        subpool = TempfilePool(self.schema, limitmb=self.limitmb, dir=self.dir)
        
        while True:
            code, args = pqueue.get()
            
            if code == -1:
                doccount = args
                break
            if code == 0:
                subpool.add_content(*args)
            elif code == 1:
                subpool.add_posting(*args)
            elif code == 2:
                subpool.add_field_length(*args)
        
        lenfilename = subpool.unique_name(".lengths")
        subpool._write_lengths(StructFile(open(lenfilename, "wb")), doccount)
        subpool.dump_run()
        rqueue.put((subpool.runs, subpool.fieldlength_totals(),
                    subpool.fieldlength_maxes(), lenfilename))


class MultiPool(PoolBase):
    def __init__(self, schema, dir=None, procs=2, limitmb=32, **kw):
        PoolBase.__init__(self, schema, dir=dir)
        
        self.procs = procs
        self.limitmb = limitmb
        
        self.postingqueue = Queue()
        self.resultsqueue = Queue()
        
        self.tasks = [PoolWritingTask(self.schema, self.dir, self.postingqueue,
                                      self.resultsqueue, self.limitmb)
                      for _ in xrange(procs)]
        for task in self.tasks:
            task.start()
    
    def add_content(self, *args):
        self.postingqueue.put((0, args))
        
    def add_posting(self, *args):
        self.postingqueue.put((1, args))
    
    def add_field_length(self, *args):
        self.postingqueue.put((2, args))
    
    def cancel(self):
        for task in self.tasks:
            task.terminate()
        self.cleanup()
    
    def cleanup(self):
        pass
    
    def finish(self, doccount, lengthfile, termtable, postingwriter):
        _fieldlength_totals = self._fieldlength_totals
        if not self.tasks:
            return
        
        pqueue = self.postingqueue
        rqueue = self.resultsqueue
        
        for _ in xrange(self.procs):
            pqueue.put((-1, doccount))
        
        #print "Joining..."
        t = now()
        for task in self.tasks:
            task.join()
        #print "Join:", now() - t
        
        #print "Getting results..."
        t = now()
        runs = []
        lenfilenames = []
        for task in self.tasks:
            taskruns, flentotals, flenmaxes, lenfilename = rqueue.get()
            runs.extend(taskruns)
            lenfilenames.append(lenfilename)
            for fieldnum, total in flentotals.iteritems():
                _fieldlength_totals[fieldnum] += total
            for fieldnum, length in flenmaxes.iteritems():
                if length > self._fieldlength_maxes.get(fieldnum, 0):
                    self._fieldlength_maxes[fieldnum] = length
        #print "Results:", now() - t
        
        #print "Writing lengths..."
        t = now()
        lw = LengthWriter(lengthfile, doccount)
        for lenfilename in lenfilenames:
            sublengths = LengthReader(StructFile(open(lenfilename, "rb")), doccount)
            lw.add_all(sublengths)
            os.remove(lenfilename)
        lw.close()
        lengths = lw.reader()
        #print "Lengths:", now() - t
        
        t = now()
        iterator = imerge([read_run(runname, count) for runname, count in runs])
        total = sum(count for runname, count in runs)
        write_postings(self.schema, termtable, lengths, postingwriter, iterator)
        for runname, count in runs:
            os.remove(runname)
        #print "Merge:", now() - t
        
        self.cleanup()
 


import unittest

from whoosh import analysis, highlight, fields, qparser
from whoosh.filedb.filestore import RamStorage


class TestHighlighting(unittest.TestCase):
    _doc = u"alfa bravo charlie delta echo foxtrot golf hotel india juliet kilo lima"
    
    def test_null_fragment(self):
        terms = frozenset(("bravo", "india"))
        sa = analysis.StandardAnalyzer()
        nf = highlight.NullFragmenter
        uc = highlight.UppercaseFormatter()
        htext = highlight.highlight(self._doc, terms, sa, nf, uc)
        self.assertEqual(htext, "alfa BRAVO charlie delta echo foxtrot golf hotel INDIA juliet kilo lima")

    def test_simple_fragment(self):
        terms = frozenset(("bravo", "india"))
        sa = analysis.StandardAnalyzer()
        sf = highlight.SimpleFragmenter(size=20)
        uc = highlight.UppercaseFormatter()
        htext = highlight.highlight(self._doc, terms, sa, sf, uc)
        self.assertEqual(htext, "alfa BRAVO charlie...hotel INDIA juliet kilo")
        
    def test_sentence_fragment(self):
        text = u"This is the first sentence. This one doesn't have the word. This sentence is the second. Third sentence here."
        terms = ("sentence", )
        sa = analysis.StandardAnalyzer(stoplist=None)
        sf = highlight.SentenceFragmenter()
        uc = highlight.UppercaseFormatter()
        htext = highlight.highlight(text, terms, sa, sf, uc)
        self.assertEqual(htext, "This is the first SENTENCE...This SENTENCE is the second...Third SENTENCE here")

    def test_context_fragment(self):
        terms = frozenset(("bravo", "india"))
        sa = analysis.StandardAnalyzer()
        cf = highlight.ContextFragmenter(terms, surround=6)
        uc = highlight.UppercaseFormatter()
        htext = highlight.highlight(self._doc, terms, sa, cf, uc)
        self.assertEqual(htext, "alfa BRAVO charlie...hotel INDIA juliet")
    
    def test_html_format(self):
        terms = frozenset(("bravo", "india"))
        sa = analysis.StandardAnalyzer()
        cf = highlight.ContextFragmenter(terms, surround=6)
        hf = highlight.HtmlFormatter()
        htext = highlight.highlight(self._doc, terms, sa, cf, hf)
        self.assertEqual(htext, 'alfa <strong class="match term0">bravo</strong> charlie...hotel <strong class="match term1">india</strong> juliet')

    def test_maxclasses(self):
        terms = frozenset(("alfa", "bravo", "charlie", "delta", "echo"))
        sa = analysis.StandardAnalyzer()
        cf = highlight.ContextFragmenter(terms, surround=6)
        hf = highlight.HtmlFormatter(tagname="b", termclass="t", maxclasses=2)
        htext = highlight.highlight(self._doc, terms, sa, cf, hf)
        self.assertEqual(htext, '<b class="match t0">alfa</b> <b class="match t1">bravo</b> <b class="match t0">charlie</b>...<b class="match t1">delta</b> <b class="match t0">echo</b> foxtrot')

    def test_workflow(self):
        st = RamStorage()
        schema = fields.Schema(id=fields.ID(stored=True),
                               title=fields.TEXT(stored=True))
        ix = st.create_index(schema)
        
        w = ix.writer()
        w.add_document(id=u"1", title=u"The man who wasn't there")
        w.add_document(id=u"2", title=u"The dog who barked at midnight")
        w.add_document(id=u"3", title=u"The invisible man")
        w.add_document(id=u"4", title=u"The girl with the dragon tattoo")
        w.add_document(id=u"5", title=u"The woman who disappeared")
        w.commit()
        
        s = ix.searcher()
        
        # Parse the user query
        parser = qparser.QueryParser("title", schema=ix.schema)
        q = parser.parse(u"man")
        
        # Extract the terms the user used in the field we're interested in
        terms = [text for fieldname, text in q.all_terms()
                 if fieldname == "title"]
        
        # Perform the search
        r = s.search(q)
        self.assertEqual(len(r), 2)
        
        # Use the same analyzer as the field uses. To be sure, you can
        # do schema[fieldname].format.analyzer. Be careful not to do this
        # on non-text field types such as DATETIME.
        analyzer = schema["title"].format.analyzer
        
        # Since we want to highlight the full title, not extract fragments,
        # we'll use NullFragmenter.
        nf = highlight.NullFragmenter
        
        # In this example we'll simply uppercase the matched terms
        fmt = highlight.UppercaseFormatter()
        
        outputs = []
        for d in r:
            text = d["title"]
            outputs.append(highlight.highlight(text, terms, analyzer, nf, fmt))
        
        self.assertEqual(outputs, ["The invisible MAN",
                                   "The MAN who wasn't there"])
        



if __name__ == '__main__':
    unittest.main()



#===============================================================================
# Copyright 2010 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import types
from array import array
from struct import Struct

try:
    from zlib import compress, decompress
    can_compress = True
except ImportError:
    can_compress = False

from whoosh.formats import Format
from whoosh.writing import PostingWriter
from whoosh.matching import Matcher, ReadTooFar
from whoosh.spans import Span
from whoosh.system import _INT_SIZE, _FLOAT_SIZE, unpack_long
from whoosh.util import utf8encode, utf8decode, length_to_byte, byte_to_length


class BlockInfo(object):
    __slots__ = ("flags", "nextoffset", "idslen", "weightslen", "postcount",
                 "maxweight", "maxwol", "minlength", "maxid", "dataoffset",
                 "_blockstart", "_pointer_pos")
    
    # On-disk header format
    # 
    # Offset  Type  Desc
    # ------  ----  -------
    # 0       B     Flags
    # 1       B     (Unused)
    # 2       H     (Unused)
    # 4       i     Delta to start of next block
    # ------------- If byte 0 == 0, the first 8 bytes are an absolute pointer
    #               to the next block (backwards compatibility)
    # 
    # 8       H     Length of the compressed IDs, or 0 if IDs are not
    #               compressed
    # 10      H     Length of the compressed weights, or 0 if the weights are
    #               not compressed
    # 12      B     Number of posts in this block
    # 13      f     Maximum weight in this block (used for quality)
    # 17      f     Maximum (weight/fieldlength) in this block (for quality)
    # 21      f     (Unused)
    # 25      B     Minimum length in this block, encoded as byte (for quality)
    #
    # Followed by either an unsigned int or string indicating the last ID in
    # this block
    
    _struct = Struct("!BBHiHHBfffB")
    
    def __init__(self, flags=None, nextoffset=None, idslen=0, weightslen=0,
                 postcount=None, maxweight=None, maxwol=None, minlength=0,
                 maxid=None, dataoffset=None):
        self.flags = flags
        self.nextoffset = nextoffset
        self.idslen = idslen
        self.weightslen = weightslen
        self.postcount = postcount
        self.maxweight = maxweight
        self.maxwol = maxwol
        self.minlength = minlength
        self.maxid = maxid
        # Position in the file where the header ends and the data begins,
        # set in from_file()
        self.dataoffset = dataoffset
        
    def __repr__(self):
        values = " ".join("%s=%r" % (name, getattr(self, name))
                          for name in self.__slots__)
        return "<%s %s>" % (self.__class__.__name, values)
        
    def to_file(self, file, stringids=False):
        flags = 1
        
        self._blockstart = file.tell()
        self._pointer_pos = self._blockstart + 4
        file.write(self._struct.pack(flags,
                                     0, 0, # unused B, H
                                     self.nextoffset,
                                     self.idslen,
                                     self.weightslen,
                                     self.postcount,
                                     self.maxweight, self.maxwol, 0,
                                     length_to_byte(self.minlength)))
        
        # Write the maximum ID after the header. We have to do this
        # separately because it might be a string (in the case of a vector)
        if stringids:
            file.write_string(utf8encode(self.maxid)[0])
        else:
            file.write_uint(self.maxid)
    
    def write_pointer(self, file):
        nextoffset = file.tell()
        file.seek(self._pointer_pos)
        file.write_int(nextoffset - self._blockstart)
        file.seek(nextoffset)
    
    @staticmethod
    def from_file(file, stringids=False):
        here = file.tell()
        
        encoded_header = file.read(BlockInfo._struct.size)
        header = BlockInfo._struct.unpack(encoded_header)
        (flags, _, _, nextoffset, idslen, weightslen, postcount, maxweight,
         maxwol, _, minlength) = header
        
        if not flags:
            nextoffset = unpack_long(encoded_header[:8])
        else:
            nextoffset = here + nextoffset
        
        assert postcount > 0
        minlength = byte_to_length(minlength)
        
        if stringids:
            maxid = utf8decode(file.read_string())[0]
        else:
            maxid = file.read_uint()
        
        dataoffset = file.tell()
        return BlockInfo(flags=flags, nextoffset=nextoffset,
                         postcount=postcount, maxweight=maxweight,
                         maxwol=maxwol, maxid=maxid, minlength=minlength,
                         dataoffset=dataoffset, idslen=idslen,
                         weightslen=weightslen)
    

class FilePostingWriter(PostingWriter):
    def __init__(self, postfile, stringids=False, blocklimit=128,
                 compressed=True, compression=3):
        self.postfile = postfile
        self.stringids = stringids

        if blocklimit > 255:
            raise ValueError("blocklimit argument must be <= 255")
        elif blocklimit < 1:
            raise ValueError("blocklimit argument must be > 0")
        self.blocklimit = blocklimit
        self.compressed = compressed
        self.compression = compression
        self.inblock = False

    def _reset_block(self):
        if self.stringids:
            self.blockids = []
        else:
            self.blockids = array("I")
        self.blockweights = array("f")
        self.blockvalues = []
        self.blocklengths = []
        self.blockoffset = self.postfile.tell()

    def start(self, format):
        if self.inblock:
            raise Exception("Called start() in a block")

        self.format = format
        self.blockcount = 0
        self.posttotal = 0
        self.startoffset = self.postfile.tell()
        
        # Magic number
        self.postfile.write_int(-48626)
        # Placeholder for block count
        self.postfile.write_uint(0)
        
        self._reset_block()
        self.inblock = True

        return self.startoffset

    def write(self, id, weight, valuestring, dfl):
        self.blockids.append(id)
        self.blockvalues.append(valuestring)
        self.blockweights.append(weight)
        if dfl:
            self.blocklengths.append(dfl)
        if len(self.blockids) >= self.blocklimit:
            self._write_block()

    def finish(self):
        if not self.inblock:
            raise Exception("Called finish() when not in a block")

        if self.blockids:
            self._write_block()

        # Seek back to the start of this list of posting blocks and writer the
        # number of blocks
        pf = self.postfile
        pf.flush()
        offset = pf.tell()
        pf.seek(self.startoffset + _INT_SIZE)
        pf.write_uint(self.blockcount)
        pf.seek(offset)
        
        self.inblock = False
        return self.posttotal

    def close(self):
        if hasattr(self, "blockids") and self.blockids:
            self.finish()
        self.postfile.close()

    def _write_block(self):
        posting_size = self.format.posting_size
        stringids = self.stringids
        pf = self.postfile
        ids = self.blockids
        values = self.blockvalues
        weights = self.blockweights
        postcount = len(ids)
        # Only compress when there are more than 4 postings in the block
        compressed = self.compressed and postcount > 4
        compression = self.compression

        if not stringids and compressed:
            compressed_ids = compress(ids.tostring(), compression)
            idslen = len(compressed_ids)
        else:
            idslen = 0
            
        if compressed:
            compressed_weights = compress(weights.tostring(), compression)
            weightslen = len(compressed_weights)
        else:
            weightslen = 0

        # TODO: THIS IS NOT CROSS-PLATFORM BECAUSE YOU ARE CONVERTING AN ARRAY
        # TO/FROM A STRING WITHOUT REGARD TO ENDIAN-NESS!!!

        # Write the blockinfo
        maxid = ids[-1]
        maxweight = max(weights)
        maxwol = 0.0
        minlength = 0
        if self.blocklengths:
            minlength = min(self.blocklengths)
            maxwol = max(w / l for w, l in zip(weights, self.blocklengths))

        blockinfo = BlockInfo(nextoffset=0, maxweight=maxweight, maxwol=maxwol,
                              minlength=minlength, postcount=postcount,
                              maxid=maxid, idslen=idslen, weightslen=weightslen)
        blockinfo.to_file(pf, stringids)
        
        # Write the IDs
        if stringids:
            for id in ids:
                pf.write_string(utf8encode(id)[0])
        elif idslen:
            pf.write(compressed_ids)
        else:
            pf.write_array(ids)
            
        # Write the weights
        if compressed:
            pf.write(compressed_weights)
        else:
            pf.write_array(weights)

        # Write the values
        if posting_size != 0:
            values_string = ""
            
            # If the size of a posting value in this format is not fixed
            # (represented by a number less than zero), write an array of value
            # lengths
            if posting_size < 0:
                lengths = array("i", (len(valuestring) for valuestring in values))
                values_string += lengths.tostring()
            
            values_string += "".join(values)
            
            if compressed:
                values_string = compress(values_string, compression)
            
            pf.write(values_string)

        # Seek back and write the pointer to the next block
        pf.flush()
        blockinfo.write_pointer(pf)
        
        self.posttotal += postcount
        self._reset_block()
        self.blockcount += 1


class FilePostingReader(Matcher):
    def __init__(self, postfile, offset, format, scorer=None,
                 fieldname=None, text=None, stringids=False):
        
        assert isinstance(offset, (int, long)), "offset is %r/%s" % (offset, type(offset))
        assert isinstance(format, Format), "format is %r/%s" % (format, type(format))
        
        self.postfile = postfile
        self.startoffset = offset
        self.format = format
        self.supports_chars = self.format.supports("characters")
        self.supports_poses = self.format.supports("positions")
        # Bind the score and quality functions to this object as methods
        
        self.scorer = scorer
        self.fieldname = fieldname
        self.text = text
        
        self.stringids = stringids
        
        assert postfile.get_int(offset) == -48626
        self.blockcount = postfile.get_uint(offset + _INT_SIZE)
        self.baseoffset = offset + _INT_SIZE * 2
        self._active = True
        self.currentblock = -1
        self._next_block()

    def __repr__(self):
        return "%s(%r, %s, %r, %r)" % (self.__class__.__name__, self.postfile,
                                       self.startoffset, self.fieldname, self.text)

    def close(self):
        pass

    def copy(self):
        return self.__class__(self.postfile, self.startoffset, self.format,
                              scorer=self.scorer, fieldname=self.fieldname,
                              text=self.text, stringids=self.stringids)

    def is_active(self):
        return self._active

    def id(self):
        return self.ids[self.i]

    def items_as(self, astype):
        decoder = self.format.decoder(astype)
        for id, value in self.all_items():
            yield (id, decoder(value))

    def supports(self, astype):
        return self.format.supports(astype)

    def value(self):
        if self.values is None: self._read_values()
        return self.values[self.i]

    def value_as(self, astype):
        decoder = self.format.decoder(astype)
        return decoder(self.value())

    def spans(self):
        if self.supports_chars:
            return [Span(pos, startchar=startchar, endchar=endchar)
                    for pos, startchar, endchar in self.value_as("characters")]
        elif self.supports_poses:
            return [Span(pos) for pos in self.value_as("positions")]
        else:
            raise Exception("Field does not support positions (%r)" % self.fieldname)

    def weight(self):
        return self.weights[self.i]
    
    def all_ids(self):
        nextoffset = self.baseoffset
        for _ in xrange(self.blockcount):
            blockinfo = self._read_blockinfo(nextoffset)
            nextoffset = blockinfo.nextoffset
            ids, __ = self._read_ids(blockinfo.dataoffset, blockinfo.postcount,
                                     blockinfo.idslen)
            for id in ids:
                yield id

    def next(self):
        if self.i == self.blockinfo.postcount - 1:
            self._next_block()
            return True
        else:
            self.i += 1
            return False

    def skip_to(self, id):
        if not self.is_active(): raise ReadTooFar
        
        i = self.i
        # If we're already in the block with the target ID, do nothing
        if id <= self.ids[i]: return
        
        # Skip to the block that would contain the target ID
        if id > self.blockinfo.maxid:
            self._skip_to_block(lambda: id > self.blockinfo.maxid)
        if not self._active: return

        # Iterate through the IDs in the block until we find or pass the
        # target
        ids = self.ids
        i = self.i
        while ids[i] < id:
            i += 1
            if i == len(ids):
                self._active = False
                return
        self.i = i

    def _read_blockinfo(self, offset):
        pf = self.postfile
        pf.seek(offset)
        return BlockInfo.from_file(pf, self.stringids)
        
    def _read_ids(self, offset, postcount, idslen):
        pf = self.postfile
        pf.seek(offset)
        
        if self.stringids:
            rs = pf.read_string
            ids = [utf8decode(rs())[0] for _ in xrange(postcount)]
            newoffset = pf.tell()
        elif idslen:
            ids = array("I")
            ids.fromstring(decompress(pf.read(idslen)))
            newoffset = offset + idslen
        else:
            ids = pf.read_array("I", postcount)
            newoffset = offset + _INT_SIZE * postcount

        return (ids, newoffset)

    def _read_weights(self, offset, postcount, weightslen):
        if weightslen:
            weights = array("f")
            weights.fromstring(decompress(self.postfile.read(weightslen)))
            newoffset = offset + weightslen
        else:
            weights = self.postfile.get_array(offset, "f", postcount)
            newoffset = offset + _FLOAT_SIZE * postcount
        return (weights, newoffset)

    def _read_values(self):
        startoffset = self.voffset
        endoffset = self.blockinfo.nextoffset
        postcount = self.blockinfo.postcount
        posting_size = self.format.posting_size

        if posting_size != 0:
            values_string = self.postfile.map[startoffset:endoffset]
            
            if self.blockinfo.weightslen:
                # Values string is compressed
                values_string = decompress(values_string)
            
            if posting_size < 0:
                # Pull the array of value lengths off the front of the string
                lengths = array("i")
                lengths.fromstring(values_string[:_INT_SIZE * postcount])
                values_string = values_string[_INT_SIZE * postcount:]
                
            # Chop up the block string into individual valuestrings
            if posting_size > 0:
                # Format has a fixed posting size, just chop up the values
                # equally
                values = [values_string[i * posting_size: i * posting_size + posting_size]
                          for i in xrange(postcount)]
            else:
                # Format has a variable posting size, use the array of lengths
                # to chop up the values.
                pos = 0
                values = []
                for length in lengths:
                    values.append(values_string[pos:pos + length])
                    pos += length
        else:
            # Format does not store values (i.e. Existence), just create fake
            # values
            values = (None,) * postcount

        self.values = values

    def _consume_block(self):
        postcount = self.blockinfo.postcount
        self.ids, woffset = self._read_ids(self.blockinfo.dataoffset, postcount,
                                           self.blockinfo.idslen)
        self.weights, voffset = self._read_weights(woffset, postcount,
                                                   self.blockinfo.weightslen)
        self.voffset = voffset
        self.values = None
        self.i = 0

    def _next_block(self, consume=True):
        self.currentblock += 1
        if self.currentblock == self.blockcount:
            self._active = False
            return

        if self.currentblock == 0:
            pos = self.baseoffset
        else:
            pos = self.blockinfo.nextoffset

        self.blockinfo = self._read_blockinfo(pos)
        if consume:
            self._consume_block()

    def _skip_to_block(self, targetfn):
        skipped = 0
        while self._active and targetfn():
            self._next_block(consume=False)
            skipped += 1

        if self._active:
            self._consume_block()
        
        return skipped
    
    def supports_quality(self):
        return self.scorer and self.scorer.supports_quality()
    
    def skip_to_quality(self, minquality):
        bq = self.block_quality
        if bq() > minquality: return 0
        return self._skip_to_block(lambda: bq() <= minquality)
    
    def score(self):
        return self.scorer.score(self)
    
    def quality(self):
        return self.scorer.quality(self)
    
    def block_quality(self):
        return self.scorer.block_quality(self)
    
    
    
    
        












